#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

"""Functions to use the Saalfeld lab's h5 transforms."""

import concurrent.futures
import h5py

import numpy as np
import pandas as pd

from scipy.interpolate import RegularGridInterpolator
from typing import Union, Optional

from .base import BaseTransform
from .affine import AffineTransform

from .. import config

logger = config.get_logger(__name__)


class H5transform(BaseTransform):
    """Hdf5 transform of 3D spatial data.

    See [here](https://github.com/saalfeldlab/template-building/wiki/Hdf5-Deformation-fields)
    for specifications of the format.

    Parameters
    ----------
    f :             str
                    Path to Hdf5 transformation.
    direction :     "forward" | "inverse"
                    Direction of transformation.
    level :         int, optional
                    For Hdf5 files with deformation fields at multiple
                    resolutions: what level of detail to use. Negative values
                    go backwards from the highest available resolution
                    (-1 = highest, -2 = second highest, etc). Ignored if only a
                    single deformation field present.
    cache :         bool
                    If True, we will cache the deformation field for subsequent
                    future transforms. This will speed up future calculations
                    in the future but comes at a memory cost.
    full_ingest :   bool
                    If True, will read and cache the full deformation field at
                    initialization. This additional upfront cost can pay off if
                    you are about to make many transforms across the volume.

    """

    def __init__(self, f: str,
                 direction: str = 'forward',
                 level: Optional[int] = -1,
                 cache: bool = False,
                 full_ingest: bool = False):
        """Init class."""
        assert direction in ('forward', 'inverse'), ('`direction` must be "forward"'
                                                     f'or "inverse", not "{direction}"')

        self.file = f
        self.direction = direction
        self.field = {'forward': 'dfield', 'inverse': 'invdfield'}[direction]

        # Trying to avoid the file repeatedly so we are making these initial
        # adjustments all in one go even though it would be more Pythonic to
        # delegate to property getter/setter methods
        with h5py.File(self.file, 'r') as h5:
            # Get the available levels
            available_levels = []
            for k in h5.keys():
                try:
                    available_levels.append(int(k))
                except ValueError:
                    continue
            available_levels = sorted(available_levels)

            # Check if there are indeed deformation fields at various resolutions
            if available_levels:
                if isinstance(level, type(None)):
                    level = available_levels[0]
                elif level < 0:
                    ix = level * -1 - 1
                    level = available_levels[ix]

                # Set level
                self._level = str(level)

                # Shape of deformation field
                self.shape = h5[self.level][self.field].shape

                # Data type of deformation field
                self.dtype = h5[self.level][self.field].dtype
            elif self.field in h5.keys():
                # Set level
                self._level = None

                # Shape of deformation field
                self.shape = h5[self.field].shape

                # Data type of deformation field
                self.dtype = h5[self.field].dtype
            else:
                raise ValueError('Unable to parse deformation fields from '
                                 f' {self.file}.')

        # Prepare cache if applicable
        if full_ingest:
            # Ingest the whole deformation field
            self.full_ingest()
        elif cache:
            self.use_cache = True

    def __eq__(self, other) -> bool:
        """Compare with other Transform."""
        if isinstance(other, H5transform):
            if self.file == other.file:
                if self.direction == other.direction:
                    if self.level == other.level:
                        return True
        return False

    def __neg__(self) -> 'H5transform':
        """Invert direction."""
        # Swap direction
        new_direction = {'forward': 'inverse',
                         'inverse': 'forward'}[self.direction]
        # We will re-iniatialize
        x = H5transform(self.file,
                        direction=new_direction,
                        level=int(self.level) if self.level else None,
                        cache=self.use_cache,
                        full_ingest=False)

        return x

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, value):
        raise ValueError('`level` cannot be changed after initialization.')

    @property
    def use_cache(self):
        """Whether to cache the deformation field."""
        if not hasattr(self, '_use_cache'):
            self._use_cache = False
        return self._use_cache

    @use_cache.setter
    def use_cache(self, value):
        """Set whether to cache the deformation field."""
        assert isinstance(value, bool)

        # If was False and now set to True, build the cache
        if not getattr(self, '_use_cache', False) and value:
            # This is the actual cache
            self.cache = np.zeros(self.shape,
                                  dtype=self.dtype)
            # This is a mask that tells us which values have already been cached
            self.cached = np.zeros(self.shape[:-1],
                                   dtype=bool)
            self._use_cache = True
        # If was True and now is set to False, deconstruct cache
        elif getattr(self, '_use_cache', False) and not value:
            del self.cache
            del self.cached
            self._use_cache = False

            if hasattr(self, '_fully_ingested'):
                del self._fully_ingested

        # Note: should explore whether we can use sparse N-dimensional
        # arrays for caching to save memory
        # See https://github.com/pydata/sparse/

    def copy(self):
        """Return copy."""
        return H5transform(self.file,
                           direction=self.direction,
                           level=int(self.level) if self.level else None,
                           cache=self.use_cache,
                           full_ingest=False)

    def full_ingest(self):
        """Fully ingest the deformation field."""
        # Skip if already ingested
        if getattr(self, '_fully_ingested', False):
            return

        with h5py.File(self.file, 'r') as h5:
            # Read in the entire field
            if self.level:
                self.cache = h5[self.level][self.field][:, :, :]
            else:
                self.cache = h5[self.field][:, :, :]
            # Keep a flag of this
            self._fully_ingested = True
            # We set `cached` to True instead of using a mask
            self.cached = True
            # Keep track of the caching
            self._use_cache = True

    def precache(self, bbox: Union[list, np.ndarray], padding=True):
        """Cache deformation field for given bounding box.

        Parameters
        ----------
        bbox :      list | array
                    Must be `[[x1, x2], [y1, y2], [z1, z2]]`.
        padding :   bool
                    If True, will add the (required!) padding to the bounding
                    box.

        """
        bbox = np.asarray(bbox)

        if bbox.ndim != 2 or bbox.shape != (3, 2):
            raise ValueError(f'Expected (3, 2) bounding box, got {bbox.shape}')

        # Set use_cache=True -> this also prepares the cache array(s)
        self.use_cache = True

        with h5py.File(self.file, 'r') as h5:
            if self.level:
                spacing = h5[self.level][self.field].attrs['spacing']
            else:
                spacing = h5[self.field].attrs['spacing']

            # Note that we invert because spacing is given in (z, y, x)
            bbox_vxl = (bbox.T / spacing[::-1]).T
            # Digitize into voxels
            bbox_vxl = bbox_vxl.round().astype(int)

            if padding:
                bbox_vxl[:, 0] -= 2
                bbox_vxl[:, 1] += 2

            # Make sure we are within bounds
            bbox_vxl = np.clip(bbox_vxl.T, 0, self.shape[:-1][::-1]).T

            # Extract values
            x1, x2, y1, y2, z1, z2 = bbox_vxl.flatten()

            # Cache values in this bounding box
            if self.level:
                self.cache[z1:z2, y1:y2, x1:x2] = h5[self.level][self.field][z1:z2, y1:y2, x1:x2]
            else:
                self.cache[z1:z2, y1:y2, x1:x2] = h5[self.field][z1:z2, y1:y2, x1:x2]
            self.cached[z1:z2, y1:y2, x1:x2] = True

    @staticmethod
    def from_file(filepath: str, **kwargs) -> 'H5transform':
        """Generate H5transform from file.

        Parameters
        ----------
        filepath :  str
                    Path to H5 transform.
        **kwargs
                    Keyword arguments passed to H5transform.__init__

        Returns
        -------
        H5transform

        """
        return H5transform(str(filepath), **kwargs)

    def xform(self,
              points: np.ndarray,
              affine_fallback: bool = True,
              force_deform: bool = True) -> np.ndarray:
        """Xform data.

        Parameters
        ----------
        points :            (N, 3) numpy array | pandas.DataFrame
                            Points to xform. DataFrame must have x/y/z columns.
        affine_fallback :   bool
                            If False, points outside the deformation field will
                            be returned as `np.nan`. If True, these points will
                            only receive the affine part of the transformation.
        force_deform :      bools
                            If True, points outside the deformation field be
                            deformed using the closest point inside the
                            deformation field. Ignored if `affine_fallback` is
                            `False`.

        Returns
        -------
        pointsxf :      (N, 3) numpy array
                        Transformed points. Points outside the deformation field
                        will have only the affine part of the transform applied.

        """
        if isinstance(points, pd.DataFrame):
            # Make sure x/y/z columns are present
            if np.any([c not in points for c in ['x', 'y', 'z']]):
                raise ValueError('points DataFrame must have x/y/z columns.')
            points = points[['x', 'y', 'z']].values

        if not isinstance(points, np.ndarray) or points.ndim != 2 or points.shape[1] != 3:
            raise TypeError('`points` must be numpy array of shape (N, 3) or '
                            'pandas DataFrame with x/y/z columns')

        # Read the file
        with h5py.File(self.file, 'r') as h5:
            if self.level:
                field = h5[self.level][self.field]
            else:
                field = h5[self.field]

            # We need the field to be (z, y, x, offsets) with `offsets` being
            # three values - for example: (293, 470, 1010, 3)
            # If that's not the case, something is fishy!
            if field.shape[-1] != 3:
                logger.warning('Expected the deformation field to be of shape '
                               f'(z, y, x, 3), got {field.shape}.')

            if 'affine' in field.attrs:
                # The affine part of the transform is a 4 x 4 matrix where the upper
                # 3 x 4 part (row x columns) is an attribute of the h5 dataset
                M = np.ones((4, 4))
                M[:3, :4] = field.attrs['affine'].reshape(3, 4)
                affine = AffineTransform(M)
            else:
                affine = False

            # Get quantization multiplier for later use
            quantization_multiplier = field.attrs.get('quantization_multiplier', 1)

            # For forward direction, the affine part is applied first
            if self.direction == 'inverse' and affine:
                xf = affine.xform(points)
            else:
                xf = points

            # Translate points into voxel space
            spacing = field.attrs['spacing']
            # Note that we invert because spacing is given in (z, y, x)
            xf_voxel = xf / spacing[::-1]
            # Digitize points into voxels
            xf_indices = xf_voxel.round().astype(int)
            # Determine the bounding box of the deformation vectors we need
            # Note that we are grabbing a bit more than required - this is
            # necessary for interpolation later down the line
            mn = xf_indices.min(axis=0) - 2
            mx = xf_indices.max(axis=0) + 2

            # Make sure we are within bounds
            # Note that we clip `mn` at 0 and `mx` at 2 at the lower end?
            # This is to make sure we have enough of the deformation field
            # to interpolate later on `offsets`
            mn = np.clip(mn, 2, np.array(self.shape[:-1][::-1])) - 2
            mx = np.clip(mx, 0, np.array(self.shape[:-1][::-1]) - 2) + 2

            # Check if we can use cached values
            if self.use_cache and (hasattr(self, '_fully_ingested')
                                   or np.all(self.cached[mn[2]: mx[2],
                                                         mn[1]: mx[1],
                                                         mn[0]: mx[0]])):
                offsets = self.cache[mn[2]: mx[2], mn[1]: mx[1], mn[0]: mx[0]]
            else:
                # Load the deformation values for this bounding box
                # This is faster than grabbing individual voxels and
                offsets = field[mn[2]: mx[2], mn[1]: mx[1], mn[0]: mx[0]]

                if self.use_cache:
                    # Write these offsets to cache
                    self.cache[mn[2]: mx[2], mn[1]: mx[1], mn[0]: mx[0]] = offsets
                    self.cached[mn[2]: mx[2], mn[1]: mx[1], mn[0]: mx[0]] = True

        # For interpolation, we need to split the offsets into their x, y
        # and z component
        xgrid = offsets[:, :, :, 0]
        ygrid = offsets[:, :, :, 1]
        zgrid = offsets[:, :, :, 2]

        xx = np.arange(mn[0], mx[0])
        yy = np.arange(mn[1], mx[1])
        zz = np.arange(mn[2], mx[2])

        # The RegularGridInterpolator is the fastest one but the results are
        # are ever so slightly (4th decimal) different from the Java impl
        xinterp = RegularGridInterpolator((zz, yy, xx), xgrid,
                                          bounds_error=False, fill_value=0)
        yinterp = RegularGridInterpolator((zz, yy, xx), ygrid,
                                          bounds_error=False, fill_value=0)
        zinterp = RegularGridInterpolator((zz, yy, xx), zgrid,
                                          bounds_error=False, fill_value=0)

        # Before we interpolate check how many points are outside the
        # deformation field -> these will only receive the affine part of the
        # transform
        is_out = (xf_voxel.min(axis=1) < 0) | np.any(xf_voxel >= self.shape[:-1][::-1], axis=1)

        # If more than 20% (arbitrary number) of voxels are out, there is
        # something suspicious going on
        frac_out = is_out.sum() / xf_voxel.shape[0]
        if frac_out > 0.2:
            logger.warning(f'A suspiciously large fraction ({frac_out:.1%}) '
                           f'of {xf_voxel.shape[0]} points appear to be outside '
                           'the H5 deformation field. Please make doubly sure '
                           'that the input coordinates are in the correct '
                           'space/units')

        # If all points are outside the volume, the interpolation complains
        if frac_out < 1 or (force_deform and affine_fallback):
            if force_deform:
                # For the purpose of finding offsets, we will snap points
                # outside the deformation field to the closest inside voxel
                q_voxel = np.clip(xf_voxel,
                                  a_min=0,
                                  a_max=np.array(self.shape[:-1][::-1]) - 1)
            else:
                q_voxel = xf_voxel

            # Interpolate coordinates and re-combine to an x/y/z array
            offset_vxl = np.vstack((xinterp(q_voxel[:, ::-1], method='linear'),
                                    yinterp(q_voxel[:, ::-1], method='linear'),
                                    zinterp(q_voxel[:, ::-1], method='linear'))).T

            # Turn offsets into real-world coordinates
            offset_real = offset_vxl * quantization_multiplier

            # Apply offsets
            # Please note that we must not use += here
            # That's to avoid running into data type errors where numpy
            # will refuse to add e.g. float64 to int64.
            # By using "+" instead of "+=" we are creating a new array that
            # is potentially upcast from e.g. int64 to float64
            xf = xf + offset_real

        # For inverse direction, the affine part is applied second
        if self.direction == 'forward' and affine:
            xf = affine.xform(xf)

        # If no affine_fallback, set outside points to np.nan
        if not affine_fallback:
            xf[is_out, :] = np.nan

        return xf


def read_points_threaded(voxels, filepath, level, dir, threads=5):
    """Some tinkering with using multiple processes to read voxels."""
    splits = np.array_split(voxels, threads)
    with concurrent.futures.ProcessPoolExecutor(max_workers=threads) as executor:
        chunks = [(arr, filepath, level, dir) for arr in splits]
        futures = executor.map(_read_points, chunks)
        offset = np.vstack(list(futures))

    return offset


def _read_points(params):
    voxels, filepath, level, dir = params
    f = h5py.File(filepath, 'r', libver='latest', swmr=True)
    # Get all these voxels
    data = []
    for vx in config.tqdm(voxels):
        data.append(f[level][dir][vx[2], vx[1], vx[0]])
    return np.vstack(data)




#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

"""Module contains functions implementing alignNBLAST."""

import os
import operator

import numpy as np
import pandas as pd
import multiprocessing as mp

from functools import partial
from typing import Callable, Dict, Union, Optional
from typing_extensions import Literal
from concurrent.futures import ProcessPoolExecutor, as_completed

from .. import config, utils, core
from ..transforms.align import align_rigid, align_deform, align_pca, _align_rigid_deform

from .base import Blaster, NestedIndices
from .nblast_funcs import (nblast_preflight,find_optimal_partition, set_omp_flag)
from .smat import Lookup2d, _nblast_v1_scoring

logger = config.logger


# This multiplier controls job size for NBLASTs (only relevant for
# multiprocessing and if progress=True).
# Larger multiplier = larger job sizes = fewer jobs = slower updates & less overhead
# Smaller multiplier = smaller job sizes = more jobs = faster updates & more overhead
JOB_SIZE_MULTIPLIER = 1


class NBlasterAlign(Blaster):
    """Implements a version of NBLAST were neurons are first aligned.

    Please note that some properties are computed on initialization and
    changing parameters (e.g. `use_alpha`) at a later stage will mess things
    up!

    Parameters
    ----------
    use_alpha :     bool
                    Whether or not to use alpha values for the scoring.
                    If True, the dotproduct of nearest neighbor vectors will
                    be scaled by `sqrt(alpha1 * alpha2)`.
    normalized :    bool
                    If True, will normalize scores by the best possible score
                    (i.e. self-self) of the query neuron.
    smat :          navis.nbl.smat.Lookup2d | pd.DataFrame | str | Callable
                    How to convert the point match pairs into an NBLAST score,
                    usually by a lookup table:
                     - if 'auto' (default), will use the "official" NBLAST scoring
                       matrices based on FCWB data. Same behaviour as in R's
                       nat.nblast implementation.
                     - if `smat='v1'` it uses the analytic formulation of the
                       NBLAST scoring from Kohl et. al (2013). You can adjust parameter
                       `sigma_scaling` (default to 10) using `smat_kwargs`.
                     - DataFrames will be used to build a `Lookup2d`
                     - if `Callable` given, it passes distance and dot products as
                       first and second argument respectively
                     - if `smat=None` the scores will be generated as the
                       product of the distances and the dotproduct of the vectors
                       of nearest-neighbor pairs
    smat_kwargs:    Dictionary with additional parameters passed to scoring
                    functions. For example: `smat_kwargs["sigma_scoring"] = 10`.
    limit_dist :    float | "auto" | None
                    Sets the max distance for the nearest neighbor search
                    (`distance_upper_bound`). Typically this should be the
                    highest distance considered by the scoring function. If
                    "auto", will extract that value from the first axis of the
                    scoring matrix.
    progress :      bool
                    If True, will show a progress bar.

    """

    def __init__(self,
                 align_func, two_way_align=True, sample_align=None,
                 use_alpha=False, normalized=True, smat='auto',
                 limit_dist=None, approx_nn=False, dtype=np.float64,
                 progress=True,
                 smat_kwargs=dict(),
                 align_kwargs=dict(),
                 dotprop_kwargs=dict(),
                 ):
        """Initialize class."""
        super().__init__(progress=progress, dtype=dtype)
        self.align_func = align_func
        self.two_way_align = two_way_align
        self.sample_align = sample_align
        self.use_alpha = use_alpha
        self.normalized = normalized
        self.approx_nn = approx_nn
        self.dotprop_kwargs = dotprop_kwargs
        self.align_kwargs = align_kwargs
        self.desc = "NBlasting"
        self.self_hits = {}
        self.dotprops = {}
        self.neurons = []

        if smat is None:
            self.score_fn = operator.mul
        elif smat == 'auto':
            from ..nbl.smat import smat_fcwb
            self.score_fn = smat_fcwb(self.use_alpha)
        elif smat == 'v1':
            self.score_fn = partial(
                _nblast_v1_scoring, sigma_scoring = smat_kwargs.get('sigma_scoring', 10)
            )
        elif isinstance(smat, pd.DataFrame):
            self.score_fn = Lookup2d.from_dataframe(smat)
        else:
            self.score_fn = smat

        if limit_dist == "auto":
            try:
                if self.score_fn.axes[0].boundaries[-1] != np.inf:
                    self.distance_upper_bound = self.score_fn.axes[0].boundaries[-1]
                else:
                    # If the right boundary is open (i.e. infinity), we will use
                    # the second highest boundary plus a 5% offset
                    self.distance_upper_bound = self.score_fn.axes[0].boundaries[-2] * 1.05
            except AttributeError:
                logger.warning("Could not infer distance upper bound from scoring function")
                self.distance_upper_bound = None
        else:
            self.distance_upper_bound = limit_dist

    def append(self, neuron: core.BaseNeuron) -> NestedIndices:
        """Append neurons.

        Returns the numerical index appended neurons.
        If neurons is a (possibly nested) sequence of neurons,
        return a (possibly nested) list of indices.

        Note that `self_hit` is ignored (and hence calculated from scratch)
        when `neurons` is a nested list of dotprops.
        """
        if isinstance(neuron, core.BaseNeuron):
            return self._append_neuron(neuron)

        try:
            return [self.append(n) for n in neuron]
        except TypeError:  # i.e. not iterable
            raise ValueError(f"Expected Neuron or iterable thereof; got {type(neuron)}")

    def _append_neuron(self, neuron: core.BaseNeuron) -> int:
        next_id = len(self)
        self.neurons.append(neuron)
        self.ids.append(neuron.id)
        return next_id

    def get_dotprop(self, ix):
        if ix not in self.dotprops:
            if not isinstance(self.neurons[ix], core.Dotprops):
                self.dotprops[ix] = core.make_dotprops(self.neurons[ix],
                                                    **self.dotprop_kwargs)
            else:
                self.dotprops[ix] = self.neurons[ix]
        return self.dotprops[ix]

    def get_self_hit(self, ix):
        if ix not in self.self_hits:
            self.self_hits[ix] = self.calc_self_hit(self.get_dotprop(ix))
        return self.self_hits[ix]

    def calc_self_hit(self, dotprops):
        """Non-normalized value for self hit."""
        if not self.use_alpha:
            return len(dotprops.points) * self.score_fn(0, 1.0)
        else:
            dists = np.repeat(0, len(dotprops.points))
            alpha = dotprops.alpha * dotprops.alpha
            dots = np.repeat(1, len(dotprops.points)) * np.sqrt(alpha)
            return self.score_fn(dists, dots).sum()

    def score_single(self, q_dp, t_dp, q_idx):
        """Calculate score for single query/target dotprop pair."""
        # Run nearest-neighbor search for query against target
        data = q_dp.dist_dots(t_dp,
                              alpha=self.use_alpha,
                              # eps=0.1 means we accept 10% inaccuracy
                              eps=.1 if self.approx_nn else 0,
                              distance_upper_bound=self.distance_upper_bound)

        if self.use_alpha:
            dists, dots, alpha = data
            dots *= np.sqrt(alpha)
        else:
            dists, dots = data

        scr = self.score_fn(dists, dots).sum()

        # Normalize against best hit
        if self.normalized:
            scr /= self.get_self_hit(q_idx)

        return scr

    def single_query_target(self, q_idx: int, t_idx: int, scores='forward',
                            allow_rev_align=True):
        """Query single query against single target."""
        # Take a short-cut if this is a self-self comparison
        if q_idx == t_idx:
            if self.normalized:
                return 1
            return self.get_self_hit(q_idx)

        # Align the query to the target
        q_xf = self.align_func(self.neurons[q_idx],
                               target=self.neurons[t_idx],
                               sample=self.sample_align,
                               progress=False,
                               **self.align_kwargs)[0][0]

        # The query must always be made into new dotprops because it has been
        # moved around
        q_dp = core.make_dotprops(q_xf, **self.dotprop_kwargs)

        # The target dotprop has to be compute only once
        t_dp = self.get_dotprop(t_idx)

        scr = self.score_single(q_dp, t_dp, q_idx)

        # For the mean score we also have to produce the reverse score
        if scores in ('mean', 'min', 'max', 'both'):
            reverse = self.score_single(t_dp, q_dp, t_idx)
            if scores == 'mean':
                scr = (scr + reverse) / 2
            elif scores == 'min':
                scr = min(scr, reverse)
            elif scores == 'max':
                scr = max(scr, reverse)
            elif scores == 'both':
                # If both scores are requested
                scr = [scr, reverse]

        if self.two_way_align and allow_rev_align:
            rev = self.single_query_target(t_idx, q_idx, scores=scores,
                                           allow_rev_align=False)
            scr = (scr + rev) / 2

        return scr


def nblast_align(query: Union[core.BaseNeuron, core.NeuronList],
                 target: Optional[str] = None,
                 align_method: Union[Literal['rigid'],
                                     Literal['deform'],
                                     Literal['pca']] = 'rigid',
                 two_way_align: bool = True,
                 sample_align: Optional[float] = None,
                 scores: Union[Literal['forward'],
                               Literal['mean'],
                               Literal['min'],
                               Literal['max']] = 'mean',
                 normalized: bool = True,
                 use_alpha: bool = False,
                 smat: Optional[Union[str, pd.DataFrame, Callable]] = 'auto',
                 limit_dist: Optional[Union[Literal['auto'], int, float]] = None,
                 approx_nn: bool = False,
                 precision: Union[int, str, np.dtype] = 64,
                 n_cores: int = os.cpu_count() // 2,
                 progress: bool = True,
                 dotprop_kwargs: Optional[Dict] = dict(),
                 align_kwargs: Optional[Dict] = dict(),
                 smat_kwargs: Optional[Dict] = dict()) -> pd.DataFrame:
    """Run NBLAST on pairwise-aligned neurons.

    Requires the `pycpd` library at least version 2.0.1 which at the time of
    writing is only available from Github (not PyPI):

      https://github.com/siavashk/pycpd

    Parameters
    ----------
    query :         Neuron | NeuronList
                    Query neuron(s) to NBLAST against the targets. Neurons
                    should be in microns as NBLAST is optimized for that and
                    have similar sampling resolutions.
    target :        Neuron | NeuronList, optional
                    Target neuron(s) to NBLAST against. Neurons should be in
                    microns as NBLAST is optimized for that and have
                    similar sampling resolutions. If not provided, will NBLAST
                    queries against themselves.
    align_method :  "rigid" | "deform" | "pca" | "rigid+deform"
                    Which method to use for alignment. Maps to the respective
                    `navis.align_{method}` function.
    two_way_align : bool
                    If True, will run the alignment + NBLAST in both,
                    query->target as well as query->target direction. This is
                    highly recommended because it reduces the chance that a
                    single bad alignment will mess up your scores.
    sample_align :  float [0-1], optional
                    If provided, will calculate an initial alignment on just a
                    fraction of the points followed by a landmark transform
                    to transform the rest. Use this to speed things up.
    scores :        'forward' | 'mean' | 'min' | 'max' | 'both'
                    Determines the final scores:

                      - 'forward' (default) returns query->target scores
                      - 'mean' returns the mean of query->target and
                        target->query scores
                      - 'min' returns the minium between query->target and
                        target->query scores
                      - 'max' returns the maximum between query->target and
                        target->query scores
                      - 'both' will return foward and reverse scores as
                        multi-index DataFrame

    use_alpha :     bool, optional
                    Emphasizes neurons' straight parts (backbone) over parts
                    that have lots of branches for the NBLAST.
    normalized :    bool, optional
                    Whether to return normalized NBLAST scores.
    smat :          str | pd.DataFrame | Callable
                    Score matrix. If 'auto' (default), will use scoring matrices
                    from FCWB. Same behaviour as in R's nat.nblast
                    implementation.
                    If `smat='v1'` it uses the analytic formulation of the
                    NBLAST scoring from Kohl et. al (2013). You can adjust parameter
                    `sigma_scaling` (default to 10) using `smat_kwargs`.
                    If `Callable` given, it passes distance and dot products as
                    first and second argument respectively.
                    If `smat=None` the scores will be
                    generated as the product of the distances and the dotproduct
                    of the vectors of nearest-neighbor pairs.
    limit_dist :    float | "auto" | None
                    Sets the max distance for the nearest neighbor search
                    (`distance_upper_bound`). Typically this should be the
                    highest distance considered by the scoring function. If
                    "auto", will extract that value from the scoring matrix.
                    While this can give a ~2X speed up, it will introduce slight
                    inaccuracies because we won't have a vector component for
                    points without a nearest neighbour within the distance
                    limits. The impact depends on the scoring function but with
                    the default FCWB `smat`, this is typically limited to the
                    third decimal (0.0086 +/- 0.0027 for an all-by-all of 1k
                    neurons).
    approx_nn :     bool
                    If True, will use approximate nearest neighbors. This gives
                    a >2X speed up but also produces only approximate scores.
                    Impact depends on the use case - testing highly recommended!
    n_cores :       int, optional
                    Max number of cores to use for nblasting. Default is
                    `os.cpu_count() // 2`. This should ideally be an even
                    number as that allows optimally splitting queries onto
                    individual processes. Also note that due to multiple layers
                    of concurrency using all available cores might not be the
                    fastest option.
    precision :     int [16, 32, 64] | str [e.g. "float64"] | np.dtype
                    Precision for scores. Defaults to 64 bit (double) floats.
                    This is useful to reduce the memory footprint for very large
                    matrices. In real-world scenarios 32 bit (single)- and
                    depending on the purpose even 16 bit (half) - are typically
                    sufficient.
    progress :      bool
                    Whether to show progress bars. This may cause some overhead,
                    so switch off if you don't really need it.
    smat_kwargs :   dict, optional
                    Dictionary with additional parameters passed to scoring
                    functions.
    align_kwargs :  dict, optional
                    Dictionary with additional parameters passed to alignment
                    function.
    dotprop_kwargs : dict, optional
                    Dictionary with additional parameters passed to
                    `navis.make_dotprops`. Only relevant if inputs aren't
                    already dotprops.


    Returns
    -------
    scores :        pandas.DataFrame
                    Matrix with NBLAST scores. Rows are query neurons, columns
                    are targets. The order is the same as in `query`/`target`
                    and the labels are based on the neurons' `.id` property.
                    Important to note that even when `q == t` and with
                    `scores=mean` the matrix will not be symmetrical because
                    we run separate alignments for the forward and the reverse
                    comparisons.

    References
    ----------
    Costa M, Manton JD, Ostrovsky AD, Prohaska S, Jefferis GS. NBLAST: Rapid,
    Sensitive Comparison of Neuronal Structure and Construction of Neuron
    Family Databases. Neuron. 2016 Jul 20;91(2):293-311.
    doi: 10.1016/j.neuron.2016.06.012.

    Examples
    --------
    >>> import navis
    >>> nl = navis.example_neurons(n=5)
    >>> nl.units
    <Quantity([8 8 8 8 8], 'nanometer')>
    >>> # Convert to microns
    >>> nl_um = nl * (8 / 1000)
    >>> # Run the align nblast
    >>> scores = navis.nblast_align(nl_um[:3], nl_um[3:],
    ...                             dotprop_kwargs=dict(k=5),
    ...                             sample_align=.2)

    See Also
    --------
    [`navis.nblast`][]
                The vanilla version of NBLAST.
    [`navis.nblast_allbyall`][]
                A more efficient way than `nblast(query=x, target=x)`.
    [`navis.nblast_smart`][]
                A smart(er) NBLAST suited for very large NBLAST.
    [`navis.synblast`][]
                A synapse-based variant of NBLAST.

    """
    if isinstance(target, type(None)):
        target = query

    # Make sure we're working on NeuronLists
    query = core.NeuronList(query)
    target = core.NeuronList(target)

    if not callable(align_method):
        align_func = {'rigid': align_rigid,
                      'deform': align_deform,
                      'pca': align_pca,
                      'rigid+deform': _align_rigid_deform}[align_method]
    else:
        align_func = align_method

    # Run NBLAST preflight checks
    nblast_preflight(query, target, n_cores,
                     req_dotprops=False,
                     req_unique_ids=True,
                     req_microns=isinstance(smat, str) and smat=='auto')

    # Find a partition that produces batches that each run in approximately
    # 10 seconds
    if n_cores and n_cores > 1:
        n_rows, n_cols = find_optimal_partition(n_cores, query, target)
        if progress:
            # If progress bar, we need to make smaller mini batches.
            # These mini jobs must not be too small - otherwise the overhead
            # from spawning and sending results between processes slows things
            # down dramatically. Here we hardcode such that we get updates
            # at most every 1%
            n_rows = max(n_rows, len(query) // 10)
            n_cols = max(n_cols, len(target) // 10)
    else:
        n_rows = n_cols = 1

    # This makes sure we don't run into multiple layers of concurrency
    # Note that it doesn't do anything for the parent process (which is great
    # if we end up not actually using multiple cores)
    with set_omp_flag(limits=1):
        # Initialize a pool of workers
        # Note that we're forcing "spawn" instead of "fork" (default on linux)!
        # This is to reduce the memory footprint since "fork" appears to inherit all
        # variables (including all neurons) while "spawn" appears to get only
        # what's required to run the job?
        with ProcessPoolExecutor(max_workers=n_cores,
                                 mp_context=mp.get_context('spawn')) as pool:
            with config.tqdm(desc='Preparing',
                             total=n_rows * n_cols,
                             leave=False,
                             disable=not progress) as pbar:
                futures = {}
                nblasters = []
                for qix in np.array_split(np.arange(len(query)), n_rows):
                    for tix in np.array_split(np.arange(len(target)), n_cols):
                        # Initialize NBlaster
                        this = NBlasterAlign(align_func=align_func,
                                             two_way_align=two_way_align,
                                             sample_align=sample_align,
                                             use_alpha=use_alpha,
                                             normalized=normalized,
                                             smat=smat,
                                             limit_dist=limit_dist,
                                             dtype=precision,
                                             approx_nn=approx_nn,
                                             progress=progress,
                                             align_kwargs=align_kwargs,
                                             dotprop_kwargs=dotprop_kwargs,
                                             smat_kwargs=smat_kwargs)

                        # Add queries and targets
                        for i, ix in enumerate(qix):
                            this.append(query[ix])
                        for i, ix in enumerate(tix):
                            this.append(target[ix])

                        # Keep track of indices of queries and targets
                        this.queries = np.arange(len(qix))
                        this.targets = np.arange(len(tix)) + len(qix)
                        this.queries_ix = qix  # this facilitates filling in the big matrix later
                        this.targets_ix = tix  # this facilitates filling in the big matrix later
                        this.pbar_position = len(nblasters) if not utils.is_jupyter() else None

                        nblasters.append(this)
                        pbar.update()

                        # If multiple cores requested, submit job to the pool right away
                        if n_cores and n_cores > 1 and (n_cols > 1 or n_rows > 1):
                            this.progress=False  # no progress bar for individual NBLASTERs
                            futures[pool.submit(this.multi_query_target,
                                                q_idx=this.queries,
                                                t_idx=this.targets,
                                                scores=scores)] = this

            # Collect results
            if futures and len(futures) > 1:
                # Prepare empty score matrix
                scores = pd.DataFrame(np.empty((len(query), len(target)),
                                               dtype=this.dtype),
                                      index=query.id, columns=target.id)
                scores.index.name = 'query'
                scores.columns.name = 'target'

                # Collect results
                # We're dropping the "N / N_total" bit from the progress bar because
                # it's not helpful here
                fmt = ('{desc}: {percentage:3.0f}%|{bar}| [{elapsed}<{remaining}]')
                for f in config.tqdm(as_completed(futures),
                                     desc='NBLASTing',
                                     bar_format=fmt,
                                     total=len(futures),
                                     smoothing=0,
                                     disable=not progress,
                                     leave=False):
                    res = f.result()
                    this = futures[f]
                    # Fill-in big score matrix
                    scores.iloc[this.queries_ix, this.targets_ix] = res.values
            else:
                scores = this.multi_query_target(this.queries,
                                                 this.targets,
                                                 scores=scores)

    return scores



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import math
import requests
import sys
import urllib

import numpy as np
import pandas as pd

from typing import Optional, Union, List, Iterable, Dict, Tuple, Any

from .. import config, core
from .eval import is_mesh
from ..transforms.templates import TemplateBrain


# Set up logging
logger = config.get_logger(__name__)


def round_smart(num: Union[int, float], prec: int = 8) -> float:
    """Round number intelligently to produce Human-readable numbers.

    This functions rounds to the Nth decimal, where N is `precision` minus the
    number of digits before the decimal. The general idea is that the bigger
    the number, the less we care about decimals - and vice versa.

    Parameters
    ----------
    num :       float | int
                A number.
    prec :      int
                The precision we are aiming for.

    Examples
    --------
    >>> import navis
    >>> navis.utils.round_smart(0.00999)
    0.00999
    >>> navis.utils.round_smart(10000000.00999)
    10000000.0

    """
    # Number of digits before decimal
    lg = math.log10(num)
    if lg < 0:
        N = 0
    else:
        N = int(lg)

    return round(num, max(prec - N, 0))


def sizeof_fmt(num, suffix='B'):
    """Bytes to Human readable."""
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def make_volume(x: Any) -> 'core.Volume':
    """Try making a navis.Volume from input object."""
    if isinstance(x, core.Volume):
        return x
    if is_mesh(x):
        inits = dict(vertices=x.vertices, faces=x.faces)
        for p in ['name', 'id', 'color']:
            if hasattr(x, p):
                inits[p] = getattr(x, p, None)
        return core.Volume(**inits)

    raise TypeError(f'Unable to coerce input of type "{type(x)}" to navis.Volume')


def is_url(x: str) -> bool:
    """Return True if str is URL.

    Examples
    --------
    >>> from navis.utils import is_url
    >>> is_url('www.google.com')
    False
    >>> is_url('http://www.google.com')
    True
    >>> is_url("ftp://download.ft-server.org:8000")
    True

    """
    parsed = urllib.parse.urlparse(x)

    if parsed.netloc and parsed.scheme:
        return True
    else:
        return False


def _type_of_script() -> str:
    """Return context (terminal, jupyter, colab, iPython) in which navis is run."""
    try:
        ipy_str = str(type(get_ipython()))  # type: ignore
        if 'zmqshell' in ipy_str:
            return 'jupyter'
        elif 'colab' in ipy_str:
            return 'colab'
        else:  # if 'terminal' in ipy_str:
            return 'ipython'
    except BaseException:
        return 'terminal'


def is_jupyter() -> bool:
    """Test if navis is run in a Jupyter notebook.

    Also returns True if inside Google colaboratory!

    Examples
    --------
    >>> from navis.utils import is_jupyter
    >>> # If run outside a Jupyter environment
    >>> is_jupyter()
    False

    """
    return _type_of_script() in ('jupyter', 'colab')


def is_blender() -> bool:
    """Test if navis is run inside Blender.

    Examples
    --------
    >>> from navis.utils import is_blender
    >>> # If run outside Blender
    >>> is_blender()
    False

    """
    return 'blender' in sys.executable.lower()


def set_loggers(level: str = 'INFO'):
    """Set levels for all associated module loggers.

    Examples
    --------
    >>> from navis.utils import set_loggers
    >>> from navis import config
    >>> # Get current level
    >>> lvl = config.logger.level
    >>> # Set new level
    >>> set_loggers('INFO')
    >>> # Revert to old level
    >>> set_loggers(lvl)

    """
    config.logger.setLevel(level)


def set_pbars(hide: Optional[bool] = None,
              leave: Optional[bool] = None,
              jupyter: Optional[bool] = None) -> None:
    """Set global progress bar behaviors.

    Parameters
    ----------
    hide :      bool, optional
                Set to True to hide all progress bars.
    leave :     bool, optional
                Set to False to clear progress bars after they have finished.
    jupyter :   bool, optional
                Set to False to force using of classic tqdm even if in
                Jupyter environment.

    Returns
    -------
    Nothing

    Examples
    --------
    >>> from navis.utils import set_pbars
    >>> # Hide progress bars after finishing
    >>> set_pbars(leave=False)
    >>> # Never show progress bars
    >>> set_pbars(hide=True)
    >>> # Never use Jupyter widget progress bars
    >>> set_pbars(jupyter=False)

    """
    if isinstance(hide, bool):
        config.pbar_hide = hide

    if isinstance(leave, bool):
        config.pbar_leave = leave

    if isinstance(jupyter, bool):
        if jupyter:
            if not is_jupyter():
                logger.error('No Jupyter environment detected.')
            else:
                config.tqdm = config.tqdm_notebook
                config.trange = config.trange_notebook
        else:
            config.tqdm = config.tqdm_classic
            config.trange = config.trange_classic

    return


def unpack_neurons(x: Union[Iterable, 'core.NeuronList', 'core.NeuronObject'],
                   raise_on_error: bool = True
                   ) -> List['core.NeuronObject']:
    """Unpack neurons and returns a list of individual neurons.

    Examples
    --------
    This is mostly for doc tests:

    >>> from navis.utils import unpack_neurons
    >>> from navis.data import example_neurons
    >>> nl = example_neurons(3)
    >>> type(nl)
    <class 'navis.core.neuronlist.NeuronList'>
    >>> # Unpack list of neuronlists
    >>> unpacked = unpack_neurons([nl, nl])
    >>> type(unpacked)
    <class 'list'>
    >>> type(unpacked[0])
    <class 'navis.core.skeleton.TreeNeuron'>
    >>> len(unpacked)
    6

    """
    neurons: list = []

    if isinstance(x, (list, np.ndarray, tuple)):
        for l in x:
            neurons += unpack_neurons(l)
    elif isinstance(x, core.BaseNeuron):
        neurons.append(x)
    elif isinstance(x, core.NeuronList):
        neurons += x.neurons
    elif raise_on_error:
        raise TypeError(f'Unknown neuron format: "{type(x)}"')

    return neurons


def set_default_connector_colors(x: Union[List[tuple], Dict[str, tuple]]
                                 ) -> None:
    """Set/update default connector colors.

    Parameters
    ----------
    x :         dict
                New default connector colors. Can be:

                   {'cn_label': (r, g, b), ..}
                   {'cn_label': {'color': (r, g, b)}, ..}

    """
    if not isinstance(x, dict):
        raise TypeError(f'Expect dict, got "{type(x)}"')

    for k, v in x.items():
        if isinstance(v, dict):
            config.default_connector_colors[k].update(v)
        else:
            config.default_connector_colors[k]['color'] = v

    return


def parse_objects(x) -> Tuple['core.NeuronList',
                              List['core.Volume'],
                              List[np.ndarray],
                              List]:
    """Categorize objects e.g. for plotting.

    Returns
    -------
    Neurons :       navis.NeuronList
    Volume :        list of navis.Volume (trimesh.Trimesh will be converted)
    Points :        list of arrays
    Visuals :       list of vispy and pygfx visuals

    Examples
    --------
    This is mostly for doc tests:

    >>> from navis.utils import parse_objects
    >>> from navis.data import example_neurons, example_volume
    >>> import numpy as np
    >>> nl = example_neurons(3)
    >>> v = example_volume('LH')
    >>> p = nl[0].nodes[['x', 'y', 'z']].values
    >>> n, vols, points, vis = parse_objects([nl, v, p])
    >>> type(n), len(n)
    (<class 'navis.core.neuronlist.NeuronList'>, 3)
    >>> type(vols), len(vols)
    (<class 'list'>, 1)
    >>> type(vols[0])
    <class 'navis.core.volumes.Volume'>
    >>> type(points), len(points)
    (<class 'list'>, 1)
    >>> type(points[0])
    <class 'numpy.ndarray'>
    >>> type(vis), len(vis)
    (<class 'list'>, 0)

    """
    # Make sure this is a list.
    if not isinstance(x, list):
        x = [x]

    # If any list in x, flatten first
    if any([isinstance(i, list) for i in x]):
        # We need to be careful to preserve order because of colors
        y = []
        for i in x:
            y += i if isinstance(i, list) else [i]
        x = y

    # Collect neuron objects, make a single NeuronList and split into types
    neurons = core.NeuronList([ob for ob in x if isinstance(ob,
                                                            (core.BaseNeuron,
                                                             core.NeuronList))],
                              make_copy=False)

    # Collect visuals
    visuals = [ob for ob in x if 'vispy' in str(type(ob)) or 'pygfx.objects' in str(type(ob))]

    # Collect and parse volumes
    volumes = [ob for ob in x if not isinstance(ob, (core.BaseNeuron,
                                                     core.NeuronList))
               and is_mesh(ob)]
    # Add templatebrains
    volumes += [ob.mesh for ob in x if isinstance(ob, TemplateBrain)]
    # Converts any non-navis meshes into Volumes
    volumes = [core.Volume(v) if not isinstance(v, core.Volume) else v for v in volumes]

    # Collect dataframes with X/Y/Z coordinates
    dataframes = [ob for ob in x if isinstance(ob, pd.DataFrame)]
    if [d for d in dataframes if False in np.isin(['x', 'y', 'z'], d.columns)]:
        logger.warning('DataFrames must have x, y and z columns.')
    # Filter to and extract x/y/z coordinates
    dataframes = [d for d in dataframes if False not in [c in d.columns for c in ['x', 'y', 'z']]]
    dataframes = [d[['x', 'y', 'z']].values for d in dataframes]

    # Collect arrays
    arrays = [ob.copy() for ob in x if isinstance(ob, np.ndarray)]
    # Remove arrays with wrong dimensions
    if [ob for ob in arrays if ob.shape[1] != 3 and ob.shape[0] != 2]:
        logger.warning('Arrays need to be of shape (N, 3) for scatter or (2, N)'
                       ' for line plots.')
    arrays = [ob for ob in arrays if any(np.isin(ob.shape, [2, 3]))]

    points = dataframes + arrays

    return neurons, volumes, points, visuals


def make_url(baseurl, *args: str, **GET) -> str:
    """Generate URL.

    Parameters
    ----------
    *args
                Will be turned into the URL. For example:

                    >>> make_url('http://neuromorpho.org', 'neuron', 'fields')
                    'http://neuromorpho.org/neuron/fields'

    **GET
                Keyword arguments are assumed to be GET request queries
                and will be encoded in the url. For example:

                    >>> make_url('http://neuromorpho.org', 'neuron', 'fields',
                    ...          page=1)
                    'http://neuromorpho.org/neuron/fields?page=1'

    Returns
    -------
    url :       str


    Examples
    --------
    >>> from navis.utils import is_url, make_url
    >>> url = make_url('http://www.google.com', 'test', query='test')
    >>> url
    'http://www.google.com/test?query=test'
    >>> is_url(url)
    True

    """
    url = baseurl
    # Generate the URL
    for arg in args:
        arg_str = str(arg)
        joiner = '' if url.endswith('/') else '/'
        relative = arg_str[1:] if arg_str.startswith('/') else arg_str
        url = requests.compat.urljoin(url + joiner, relative)
    if GET:
        url += f'?{urllib.parse.urlencode(GET)}'
    return url


def check_vispy():
    """Check that vispy works.

    Returns
    -------
    vispy.Viewer
        A viewer which can be closed after use.

    Examples
    --------
    >>> import navis
    >>> viewer = navis.utils.check_vispy()
    >>> # When the viewer and neurons show up...
    >>> navis.close3d()
    """
    from ..data import example_neurons
    nl = example_neurons()
    return nl.plot3d(backend='vispy')



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

"""Functions to use the Saalfeld lab's h5 transforms. Requires jpype."""

import os

import numpy as np
import pandas as pd

from .base import BaseTransform, trigger_init

from .. import config

# jpype is a soft dependency - defer import errors until we first try to use it
try:
    import jpype
except ModuleNotFoundError as e:
    jpype = None
    jpype_import_error = e


# Path for the compiled transform-helpers jar which contains the required classes
fp = os.path.dirname(__file__)
jar_path = os.path.join(fp, 'jars/*')


class H5JavaTransform(BaseTransform):
    """Hdf5 transform using the Saalfeld lab's Java implementation.

    This was written mostly for validation of the pure Python implementation.
    Since it requires the compiled java code in ./jars/ which is not packaged
    with the wheel, it will only work if you cloned navis' Github repository.

    Requires `jpype`:

        pip3 install JPype1

    See [here](https://github.com/saalfeldlab/template-building/wiki/Hdf5-Deformation-fields)
    for specifications of the format.

    Parameters
    ----------
    regs :          str
                    Path to h5 transformation.
    direction :     "forward" | "inverse"
                    Direction of transformation.
    level :         int
                    What level of detail to use. Negative values default to the
                    highest available resolution.

    """

    def __init__(self, reg, direction='forward', level=-1):
        """Init class."""
        assert direction in ('forward', 'inverse'), ('`direction` must be "fowrard"'
                                                     f'or "inverse", not "{direction}"')

        self.reg = reg
        self.direction = direction
        self.level = level
        self.initialized = False

    def __delayed_init__(self, ):
        """Delayed initialization."""
        # Check if jpype is present...
        if not jpype:
            # ... else raise import exception
            raise jpype_import_error

        # Check if JVM is running
        if not jpype.isJVMStarted():
            # This takes ~1 sec and we should not do this repeatedly
            # Also note that we NEED to give it the path to the
            # compiled jar
            jpype.startJVM(classpath=[jar_path])

        # This must be set before we start accessing/setting other attributes
        # that require the transform to be initialized - to prevent infinite
        # loops
        self.initialized = True

        # Import java classes
        # -> these are imported from the transform-helpers.jar
        # -> note that for some reason the "from ch. .... import" does not work
        HDF5Factory = jpype.JPackage("ch").systemsx.cisd.hdf5.HDF5Factory
        N5HDF5Reader = jpype.JPackage("org").janelia.saalfeldlab.n5.hdf5.N5HDF5Reader
        N5DisplacementField = jpype.JPackage("org").janelia.saalfeldlab.n5.imglib2.N5DisplacementField

        # Open the h5 stack
        self.hdf5Reader = HDF5Factory.openForReading(self.reg)

        # Generate n5 reader
        self.n5 = N5HDF5Reader(self.hdf5Reader, [16, 16, 16])

        # Parse/Check level
        if self.level < 0:
            self.level = self.available_levels[0]
        elif not self.n5.exists(f'/{self.level}'):
            raise ValueError(f'`level` {self.level} does not exists in n5 stack')

        # Generate transform
        path = f'/{self.level}'
        path += '/invdfield' if self.direction == 'inverse' else '/dfield'
        self.transform = N5DisplacementField.open(self.n5, path,
                                                  self.direction == 'inverse')

    def __eq__(self, other):
        """Compare with other Transform."""
        if isinstance(other, H5JavaTransform):
            if self.reg == other.reg:
                if self.direction == other.direction:
                    if self.level == other.level:
                        return True
        return False

    def __neg__(self):
        """Invert direction."""
        # Swap direction
        new_direction = {'forward': 'inverse',
                         'inverse': 'forward'}[self.direction]
        # We will re-iniatialize
        x = H5JavaTransform(self.reg, direction=new_direction, level=self.level)

        return x

    @property
    @trigger_init
    def available_levels(self):
        """Check which levels exists in h5 stack."""
        levels = []
        for i in range(0, 10):
            if self.n5.exists(f'/{i}'):
                levels.append(i)
        return levels

    def copy(self):
        """Return copy."""
        return H5JavaTransform(self.reg,
                               direction=self.direction,
                               level=int(self.level))

    @staticmethod
    def from_file(filepath, **kwargs):
        """Generate H5transform from file.

        Parameters
        ----------
        filepath :  str
                    Path to H5 transform.
        **kwargs
                    Keyword arguments passed to H5transform.__init__

        Returns
        -------
        H5transform

        """
        defaults = {'direction': 'forward'}
        defaults.update(kwargs)
        return H5JavaTransform(str(filepath), **defaults)

    @trigger_init
    def xform(self, points, progress=False):
        """Xform data.

        Parameters
        ----------
        points :        (N, 3) numpy array | pandas.DataFrame
                        Points to xform. DataFrame must have x/y/z columns.
        progress :      bool
                        Whether to show a progress bar.

        Returns
        -------
        pointsxf :      (N, 3) numpy array
                        Transformed points. Points that failed to transform will
                        be `np.nan`.

        """
        if isinstance(points, pd.DataFrame):
            # Make sure x/y/z columns are present
            if np.any([c not in points for c in ['x', 'y', 'z']]):
                raise ValueError('points DataFrame must have x/y/z columns.')
            points = points[['x', 'y', 'z']].values

        if not isinstance(points, np.ndarray) or points.ndim != 2 or points.shape[1] != 3:
            raise TypeError('`points` must be numpy array of shape (N, 3) or '
                            'pandas DataFrame with x/y/z columns')

        # An [x, y, z] array for Java to write into
        q = jpype.JArray(jpype.JFloat, 1)([0, 0, 0])

        # It looks like we have to xform one point at a time
        xf = []
        for p in config.tqdm(points,
                             disable=not progress,
                             leave=False,
                             desc='Xforming'):
            # Xform this point
            self.transform.apply(p, q)
            # Keep track of new coordinates before moving to the next
            xf.append(list(q))

        return np.array(xf)



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import numpy as np

from scipy.ndimage import gaussian_filter
from typing import Union

from .. import core, utils

NeuronObject = Union["core.NeuronList", "core.TreeNeuron"]

__all__ = ["smooth_voxels", "thin_voxels"]


@utils.map_neuronlist(desc="Smoothing", allow_parallel=True)
def smooth_voxels(
    x: NeuronObject, sigma: int = 1, inplace: bool = False
) -> NeuronObject:
    """Smooth voxel(s) using a Gaussian filter.

    Parameters
    ----------
    x :             TreeNeuron | NeuronList
                    Neuron(s) to be processed.
    sigma :         int | (3, ) ints, optional
                    Standard deviation for Gaussian kernel. The standard
                    deviations of the Gaussian filter are given for each axis
                    as a sequence, or as a single number, in which case it is
                    equal for all axes.
    inplace :       bool, optional
                    If False, will use and return copy of original neuron(s).

    Returns
    -------
    VoxelNeuron/List
                    Smoothed neuron(s).

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1, kind='mesh')
    >>> vx = navis.voxelize(n, pitch='1 micron')
    >>> smoothed = navis.smooth_voxels(vx, sigma=2)

    See Also
    --------
    [`navis.smooth_mesh`][]
                    For smoothing MeshNeurons and other mesh-likes.
    [`navis.smooth_skeleton`][]
                    For smoothing TreeNeurons.

    """
    # The decorator makes sure that at this point we have single neurons
    if not isinstance(x, core.VoxelNeuron):
        raise TypeError(f"Can only process VoxelNeurons, not {type(x)}")

    if not inplace:
        x = x.copy()

    # Apply gaussian
    x._data = gaussian_filter(x.grid.astype(np.float32), sigma=sigma)
    x._clear_temp_attr()

    return x


@utils.map_neuronlist(desc="Thinning", allow_parallel=True)
def thin_voxels(x, inplace=False):
    """Skeletonize image data to single voxel width.

    This is a simple thin wrapper around scikit-learn's `skeletonize`.

    Parameters
    ----------
    x :         VoxelNeuron | numpy array
                The image to thin.
    inplace :   bool
                For VoxelNeurons only: Whether to manipulate the neuron
                in place.

    Returns
    -------
    thin
                Thinned VoxelNeuron or numpy array.

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1, kind='mesh')
    >>> vx = navis.voxelize(n, pitch='1 micron')
    >>> thinned = navis.thin_voxels(vx)

    """
    try:
        from skimage.morphology import skeletonize
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "`thin_image` requires the scikit-image packge:\n"
            "  pip install scikit-image"
        )

    if isinstance(x, core.VoxelNeuron):
        if not inplace:
            x = x.copy()

        x.grid = skeletonize(x.grid)
    elif isinstance(x, np.ndarray):
        x = skeletonize(x)
    else:
        raise TypeError(f"Unable to thin data of type {type(x)}")

    return x



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

from concurrent.futures import ThreadPoolExecutor
import os
import random
import re
import types
import uuid

import networkx as nx

import numpy as np
import pandas as pd

from pathlib import Path
from typing import (Sequence, Union, Iterable, List,
                    Optional, Callable, Iterator)

from .. import utils, config, core

__all__ = ['NeuronList']

# Set up logging
logger = config.get_logger(__name__)

# Max number of neurons in NeuronList to search for attributes
# This is to avoid searching through a huge list of neurons
# and freezing the interpreter
MAX_SEARCH = 100


class NeuronList:
    """Collection of neurons.

    Gives quick access to neurons' attributes and functions.

    Parameters
    ----------
    x :                 list | array | TreeNeuron | MeshNeuron | Dotprops | NeuronList
                        Data to construct neuronlist from. Can be either:

                        1. Tree/MeshNeuron(s) or Dotprops
                        2. NeuronList(s)
                        3. Anything that constructs a Tree/MeshNeuron
                        4. List of the above

    make_copy :         bool, optional
                        If True, Neurons are deepcopied before being
                        assigned to the NeuronList.
    make_using :        function | class, optional
                        Function or class used to construct neurons from
                        elements in `x` if they aren't already neurons.
                        By default, will use `navis.Neuron` to try to infer
                        what kind of neuron can be constructed.
    parallel :          bool
                        If True, will use parallel threads when initialising the
                        NeuronList. Should be slightly up to a lot faster
                        depending on the numbers of cores and the input data.
    n_cores :           int
                        Number of cores to use for when `parallel=True`.
                        Defaults to half the available cores.
    **kwargs
                        Will be passed to constructor of Tree/MeshNeuron (see
                        `make_using`).

    """

    neurons: List['core.NeuronObject']

    cable_length: Sequence[float]

    soma: Sequence[int]
    root: Sequence[int]

    graph: 'nx.DiGraph'
    igraph: 'igraph.Graph'  # type: ignore  # doesn't know iGraph

    def __init__(self,
                 x: Union[Iterable[Union[core.BaseNeuron,
                                         'NeuronList',
                                         pd.DataFrame]],
                          'NeuronList',
                          core.BaseNeuron,
                          pd.DataFrame],
                 make_copy: bool = False,
                 make_using: Optional[type] = None,
                 parallel: bool = False,
                 n_cores: int = os.cpu_count() // 2,
                 **kwargs):
        # If below parameter is True, most calculations will be parallelized
        # which speeds them up quite a bit. Unfortunately, this uses A TON of
        # memory - for large lists this might make your system run out of
        # memory. In these cases, leave this property at False
        self.parallel = parallel
        self.n_cores = n_cores

        # Determines if subsetting this NeuronList will copy the neurons
        self.copy_on_subset: bool = False

        if isinstance(x, NeuronList):
            # We can't simply say self.neurons = x.neurons b/c that way
            # changes in the list would backpropagate
            self.neurons = [n for n in x.neurons]
        elif utils.is_iterable(x):
            # If x is a list of mixed objects we need to unpack/flatten that
            # E.g. x = [NeuronList, NeuronList, core.TreeNeuron]
            # We need to make sure the order is retained though (important for
            # e.g. plotting)
            self.neurons = []
            for n in x:
                # Unpack neuronlists
                if isinstance(n, NeuronList):
                    self.neurons += n.neurons
                # Everything else is just appended - will throw error later
                else:
                    self.neurons.append(n)
        elif isinstance(x, type(None)):
            # Empty Neuronlist
            self.neurons = []
        else:
            # Any other datatype will simply be assumed to be accepted by
            # core.Neuron() - if not this will throw an error
            self.neurons = [x]  # type: ignore

        # Now convert and/or make copies if necessary
        to_convert = []
        for i, n in enumerate(self.neurons):
            if not isinstance(n, core.BaseNeuron) or make_copy is True:
                # The `i` keeps track of the original index so that after
                # conversion to Neurons, the objects will occupy the same
                # position
                to_convert.append((n, i))

        if to_convert:
            if not make_using:
                make_using = core.Neuron
            elif not isinstance(make_using, type) and not callable(make_using):
                make_using = make_using.__class__

            if self.parallel:
                with ThreadPoolExecutor(max_workers=self.n_cores) as e:
                    futures = e.map(lambda x: make_using(x, **kwargs),
                                    [n[0] for n in to_convert])

                    converted = [n for n in config.tqdm(futures,
                                                        total=len(to_convert),
                                                        desc='Make nrn',
                                                        disable=config.pbar_hide,
                                                        leave=config.pbar_leave)]

                    for i, c in enumerate(to_convert):
                        self.neurons[c[1]] = converted[i]

            else:
                for n in config.tqdm(to_convert, desc='Make nrn',
                                     disable=config.pbar_hide or len(to_convert) == 1,
                                     leave=config.pbar_leave):
                    self.neurons[n[1]] = make_using(n[0], **kwargs)

        # Add ID-based indexer
        self.idx = _IdIndexer(self)

    @property
    def neurons(self):
        """Neurons contained in this NeuronList."""
        return self.__dict__.get('neurons', [])

    @property
    def is_mixed(self):
        """Return True if contains more than one type of neuron."""
        return len(self.types) > 1

    @property
    def is_degenerated(self):
        """Return True if contains neurons with non-unique IDs."""
        return len(set(self.id)) < len(self.neurons)

    @property
    def types(self):
        """Return neuron types present in this list."""
        return tuple(set([type(n) for n in self.neurons]))

    @property
    def shape(self):
        """Shape of NeuronList (N, )."""
        return (self.__len__(),)

    @property
    def bbox(self):
        """Bounding box across all neurons in the list."""
        if self.empty:
            raise ValueError('No bounding box - NeuronList is empty.')

        bboxes = np.hstack([n.bbox for n in self.neurons])
        mn = np.min(bboxes, axis=1)
        mx = np.max(bboxes, axis=1)
        return np.vstack((mn, mx)).T

    @property
    def empty(self):
        """Return True if NeuronList is empty."""
        return len(self.neurons) == 0

    def __reprframe__(self):
        """Return truncated DataFrame for self representation."""
        if self.empty:
            return pd.DataFrame([])
        elif len(self) < 5:
            return self.summary()
        else:
            nl = self[:3] + self[-3:]
            s = nl.summary()
            # Fix index
            s.index = np.append(s.index[:3], np.arange(len(self)-3, len(self)))
            return s

    def __reprheader__(self, html=False):
        """Generate header for representation."""
        if len(self) <= 2000:
            size = utils.sizeof_fmt(self.memory_usage(deep=False, estimate=True))
            head = f'{type(self)} containing {len(self)} neurons ({size})'
        else:
            # For larger lists, extrapolate from sampling 10% of the list
            size = utils.sizeof_fmt(self.memory_usage(deep=False,
                                                      sample=True,
                                                      estimate=True))
            head = f'{type(self)} containing {len(self)} neurons (est. {size})'

        if html:
            head = head.replace('<', '&lt;').replace('>', '&gt;')

        return head

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        string = self.__reprheader__(html=False)
        if not self.empty:
            with pd.option_context("display.max_rows", 4,
                                   "display.show_dimensions", False):
                string += f'\n{str(self.__reprframe__())}'

        return string

    def _repr_html_(self):
        string = self.__reprheader__(html=True)
        if not self.empty:
            with pd.option_context("display.max_rows", 4,
                                   "display.show_dimensions", False):
                string += self.__reprframe__()._repr_html_()
        return string

    def __iter__(self) -> Iterator['core.NeuronObject']:
        """Iterator instanciates a new class every time it is called.
        This allows the use of nested loops on the same NeuronList object.
        """
        class prange_iter(Iterator['core.NeuronObject']):
            def __init__(self, neurons, start):
                self.iter = start
                self.neurons = neurons

            def __next__(self) -> 'core.NeuronObject':
                if self.iter >= len(self.neurons):
                    raise StopIteration
                to_return = self.neurons[self.iter]
                self.iter += 1
                return to_return

        return prange_iter(self.neurons, 0)

    def __len__(self):
        """Number of neurons in this list."""
        return len(self.neurons)

    def __dir__(self):
        """Custom __dir__ to add some parameters that we want to make searchable."""
        add_attr = set.union(*[set(dir(n)) for n in self.neurons[:MAX_SEARCH]])

        return list(set(super().__dir__() + list(add_attr)))

    def __getattr__(self, key):
        if self.empty:
            raise AttributeError(f'Neuronlist is empty - "{key}" not found')
        # Dynamically check if the requested attribute/function exists in
        # all neurons
        values = [getattr(n, key, NotImplemented) for n in self.neurons]
        is_method = [isinstance(v, types.MethodType) for v in values]
        # is_none = [isinstance(v, type(None)) for v in values]
        is_frame = [isinstance(v, pd.DataFrame) for v in values]
        is_quantity = [isinstance(v, config.ureg.Quantity) for v in values]

        # First check if there is any reason why we can't collect this
        # attribute across all neurons
        if all([isinstance(v, type(NotImplemented)) for v in values]):
            raise AttributeError(f'Attribute "{key}" not found in '
                                 'NeuronList or its neurons')
        elif any([isinstance(v, type(NotImplemented)) for v in values]):
            raise AttributeError(f'Attribute or function "{key}" missing '
                                 'for some neurons')
        elif len(set(is_method)) > 1:
            raise TypeError('Found both methods and attributes with name '
                            f'"{key}" among neurons.')
        # Concatenate if dealing with DataFrame
        elif not all(is_method):
            if any(is_frame):
                df = pd.concat([v for v in values if isinstance(v, pd.DataFrame)],
                               axis=0,
                               ignore_index=True,
                               join='outer',
                               sort=True)

                # For each row label which neuron (id) it belongs to
                df['neuron'] = None
                ix = 0
                for k, v in enumerate(values):
                    if isinstance(v, pd.DataFrame):
                        df.iloc[ix:ix + v.shape[0],
                                df.columns.get_loc('neuron')] = self.neurons[k].id
                        ix += v.shape[0]
                return df
            elif all(is_quantity):
                # See if units are all compatible
                is_compatible = [values[0].is_compatible_with(v) for v in values]
                if all(is_compatible):
                    # Convert all to the same units
                    conv = [v.to(values[0]).magnitude for v in values]
                    # Return pint array
                    return config.ureg.Quantity(np.array(conv), values[0].units)
                else:
                    logger.warning(f'"{key}" contains incompatible units. '
                                   'Returning unitless values.')
                    return np.array([v.magnitude for v in values])
            elif any(is_quantity):
                logger.warning(f'"{key}" contains data with and without '
                               'units. Removing units.')
                return np.array([getattr(v, 'magnitude', v) for v in values])
            else:
                # If the result would be a ragged array specify dtype as object
                # This avoids a depcrecation warning and future issues
                dtype = None
                if any([utils.is_iterable(v) for v in values]):
                    if not all([utils.is_iterable(v) for v in values]):
                        dtype = object
                    elif len(set([len(v) for v in values])) > 1:
                        dtype = object
                return np.array(values, dtype=dtype)
        # If everything is a method
        else:
            # To avoid confusion we will not allow calling of magic methods
            # via the NeuronProcessor as those are generally expected to
            # be methods of the NeuronList itself
            if key.startswith('__') and key.endswith('__'):
                raise AttributeError(f"'NeuronList' object has no attribute '{key}'")

            # Delayed import to avoid circular import
            from .core_utils import NeuronProcessor

            # Return function but wrap it in a function that will show
            # a progress bar. Note that we do not use parallel processing by
            # default to avoid errors with `inplace=True`
            return NeuronProcessor(self,
                                   values,
                                   parallel=False,
                                   desc=key)

    def __setattr__(self, key, value):
        # We have cater for the situation when we want to replace the whole
        # dictionary - e.g. when unpickling (see __setstate__)
        # Below code for setting the dictionary looks complicated and
        # unnecessary but is really complicated and VERY necessary
        if key == '__dict__':
            if not isinstance(value, dict):
                raise TypeError(f'__dict__ must be dict, got {type(value)}')
            self.__dict__.clear()
            for k, v in value.items():
                self.__dict__[k] = v
            return

        # Check if this attribute exists in the neurons
        if any([hasattr(n, key) for n in self.neurons]):
            logger.warning('It looks like you are trying to add a neuron '
                           'attribute to a NeuronList. Setting the attribute '
                           f'"{key}" on the NeuronList will not propagated to '
                           'the neurons it contains! To set neuron attributes '
                           'use the `NeuronList.set_neuron_attributes()` method.')

        self.__dict__[key] = value

    def __getstate__(self):
        """Get state (used e.g. for pickling)."""
        # We have to implement this to make sure that we don't accidentally
        # call __getstate__ of each neuron via the NeuronProcessor
        state = {k: v for k, v in self.__dict__.items() if not callable(v)}
        return state

    def __setstate__(self, d):
        """Set state (used e.g. for unpickling)."""
        # We have to implement this to make sure that we don't accidentally
        # call __setstate__ of each neuron via the NeuronProcessor
        self.__dict__ = d

    def __contains__(self, x):
        return x in self.neurons

    def __copy__(self):
        return self.copy(deepcopy=False)

    def __deepcopy__(self):
        return self.copy(deepcopy=True)

    def __getitem__(self, key):
        if utils.is_iterable(key):
            if all([isinstance(k, (bool, np.bool_)) for k in key]):
                if len(key) != len(self.neurons):
                    raise IndexError('boolean index did not match indexed '
                                     f'NeuronList; dimension is {len(self.neurons)} '
                                     'but corresponding boolean dimension is '
                                     f'{len(key)}')
                subset = [n for i, n in enumerate(self.neurons) if key[i]]
            else:
                subset = [self[i] for i in key]
        elif isinstance(key, str):
            subset = [n for n in self.neurons if re.fullmatch(key, getattr(n, 'name', ''))]

            # For indexing by name, we expect a match
            if not subset:
                raise AttributeError('NeuronList does not contain neuron(s) '
                                     f'with name: "{key}"')

        elif isinstance(key, (int, np.integer, slice)):
            subset = self.neurons[key]
        else:
            raise NotImplementedError(f'Indexing NeuronList by {type(key)} not implemented')

        if isinstance(subset, core.BaseNeuron):
            return subset

        # Make sure we unpack neurons
        subset = utils.unpack_neurons(subset)

        return self.__class__(subset, make_copy=self.copy_on_subset)

    def __setitem__(self, key, value):
        if isinstance(key, str):
            if not utils.is_iterable(value):
                for n in self.neurons:
                    setattr(n, key, value)
            elif len(value) == len(self.neurons):
                for n, v in zip(self.neurons, value):
                    setattr(n, key, v)
            else:
                raise ValueError('Length of values does not match number of '
                                 'neurons in NeuronList.')
        else:
            msg = ('Itemsetter can only be used to set attributes of the '
                   'neurons contained in the NeuronList. For example:\n'
                   '  >>> nl = navis.example_neurons(3)\n'
                   '  >>> nl["propertyA"] = 1\n'
                   '  >>> nl[0].propertyA\n'
                   '  1\n'
                   '  >>> nl["propertyB"] = ["a", "b", "c"]\n'
                   '  >>> nl[2].propertyB\n'
                   '  "c"')
            raise NotImplementedError(msg)

    def __missing__(self, key):
        raise AttributeError('No neuron matching the search criteria.')

    def __add__(self, to_add):
        """Implement addition."""
        if isinstance(to_add, core.BaseNeuron):
            return self.__class__(self.neurons + [to_add],
                                  make_copy=self.copy_on_subset)
        elif isinstance(to_add, NeuronList):
            return self.__class__(self.neurons + to_add.neurons,
                                  make_copy=self.copy_on_subset)
        elif utils.is_iterable(to_add):
            if False not in [isinstance(n, core.BaseNeuron) for n in to_add]:
                return self.__class__(self.neurons + list(to_add),
                                      make_copy=self.copy_on_subset)
            else:
                return self.__class__(self.neurons + [core.BaseNeuron[n] for n in to_add],
                                      make_copy=self.copy_on_subset)
        else:
            return NotImplemented

    def __eq__(self, other):
        """Implement equality."""
        if isinstance(other, NeuronList):
            if len(self) != len(other):
                return False
            else:
                return all([n1 == n2 for n1, n2 in zip(self, other)])
        else:
            return NotImplemented

    def __sub__(self, to_sub):
        """Implement substraction."""
        if isinstance(to_sub, core.BaseNeuron):
            return self.__class__([n for n in self.neurons if n != to_sub],
                                  make_copy=self.copy_on_subset)
        elif isinstance(to_sub, NeuronList):
            return self.__class__([n for n in self.neurons if n not in to_sub],
                                  make_copy=self.copy_on_subset)
        else:
            return NotImplemented

    def __truediv__(self, other):
        """Implements division for coordinates (nodes, connectors)."""
        return self.__class__([n / other for n in config.tqdm(self.neurons,
                                                              desc='Dividing',
                                                              disable=config.pbar_hide,
                                                              leave=False)])


    def __mul__(self, other):
        """Implement multiplication for coordinates (nodes, connectors)."""
        return self.__class__([n * other for n in config.tqdm(self.neurons,
                                                              desc='Multiplying',
                                                              disable=config.pbar_hide,
                                                              leave=False)])

    def __and__(self, other):
        """Implement bitwise AND using the & operator."""
        if isinstance(other, core.BaseNeuron):
            return self.__class__([n for n in self.neurons if n == other],
                                  make_copy=self.copy_on_subset)
        elif isinstance(other, NeuronList):
            return self.__class__([n for n in self.neurons if n in other],
                                  make_copy=self.copy_on_subset)
        else:
            return NotImplemented

    def __or__(self, other):
        """Implement bitwise OR using the | operator."""
        if isinstance(other, core.BaseNeuron):
            neurons = self.neurons
            if not any(n == other for n in neurons):
                neurons.append(other)
            return self.__class__(neurons, make_copy=self.copy_on_subset)
        elif isinstance(other, NeuronList):
            neurons = self.neurons + [n for n in other.neurons if n not in self]
            return self.__class__(neurons, make_copy=self.copy_on_subset)
        else:
            return NotImplemented

    def append(self, v):
        """Add neuron(s) to this list.

        Examples
        --------
        >>> import navis
        >>> # This is mostly for doctests
        >>> nl = navis.example_neurons()
        >>> len(nl)
        5
        >>> # Add a single neuron to the list
        >>> nl.append(nl[0])
        >>> len(nl)
        6
        >>> # Add a list of neurons to the list
        >>> nl.append(nl)
        >>> len(nl)
        12

        """
        if isinstance(v, core.BaseNeuron):
            self.neurons.append(v)
        elif isinstance(v, NeuronList):
            self.neurons += v.neurons
        else:
            raise NotImplementedError('Unable to append data of type'
                                      f'{type(v)} to NeuronList')

    def apply(self,
              func: Callable,
              *,
              parallel: bool = False,
              n_cores: int = os.cpu_count() // 2,
              omit_failures: bool = False,
              **kwargs):
        """Apply function across all neurons in this NeuronList.

        Parameters
        ----------
        func :          callable
                        Function to be applied. Must accept
                        [`navis.BaseNeuron`][] as first argument.
        parallel :      bool
                        If True (default) will use multiprocessing. Spawning the
                        processes takes time (and memory). Using `parallel=True`
                        makes only sense if the NeuronList is large or the
                        function takes a long time to run.
        n_cores :       int
                        Number of CPUs to use for multiprocessing. Defaults to
                        half the available cores.
        omit_failures : bool
                        If True, will ignore failures.

        **kwargs
                    Will be passed to function.

        Returns
        -------
        Results

        Examples
        --------
        >>> import navis
        >>> nl = navis.example_neurons()
        >>> # Apply resampling function
        >>> nl_rs = nl.apply(navis.resample_skeleton, resample_to=1000, inplace=False)

        """
        if not callable(func):
            raise TypeError('"func" must be callable')

        # Delayed import to avoid circular import
        from .core_utils import NeuronProcessor
        proc = NeuronProcessor(self,
                               func,
                               parallel=parallel,
                               n_cores=n_cores,
                               omit_failures=omit_failures,
                               desc=f'Apply {func.__name__}')

        return proc(self.neurons, **kwargs)

    def sum(self) -> pd.DataFrame:
        """Return sum numeric and boolean values over all neurons."""
        return self.summary().sum(numeric_only=True)

    def mean(self) -> pd.DataFrame:
        """Return mean numeric and boolean values over all neurons."""
        return self.summary().mean(numeric_only=True)

    def memory_usage(self, deep=False, estimate=False, sample=False):
        """Return estimated size in memory of this NeuronList.

        Works by going over each neuron and summing up their size in memory.

        Parameters
        ----------
        deep :          bool
                        Pass to pandas DataFrames. If True will inspect data of
                        object type too.
        estimate :      bool
                        If True, we will only estimate the size. This is
                        considerably faster but will slightly underestimate the
                        memory usage.
        sample :        bool
                        If True, we will only sample 10% of the neurons
                        contained in the list and extrapolate an estimate from
                        there.

        Returns
        -------
        int
                    Memory usage in bytes.

        """
        if self.empty:
            return 0

        if not sample:
            try:
                return sum([n.memory_usage(deep=deep,
                                           estimate=estimate) for n in self.neurons])
            except BaseException:
                return 0
        else:
            try:
                s = sum([n.memory_usage(deep=deep,
                                        estimate=estimate) for n in self.neurons[::10]])
                return s * (len(self.neurons) / len(self.neurons[::10]))
            except BaseException:
                return 0

    def sample(self, N: Union[int, float] = 1) -> 'NeuronList':
        """Return random subset of neurons."""
        if N < 1 and N > 0:
            N = int(len(self.neurons) * N)

        indices = list(range(len(self.neurons)))
        random.shuffle(indices)
        return self.__class__([n for i, n in enumerate(self.neurons) if i in indices[:N]],
                              make_copy=self.copy_on_subset)

    def plot3d(self, **kwargs):
        """Plot neuron in 3D using [`navis.plot3d`][].

        Parameters
        ----------
        **kwargs
                Keyword arguments will be passed to [`navis.plot3d`][].
                See `help(navis.plot3d)` for a list of keywords.

        See Also
        --------
        [`navis.plot3d`][]
                Base function called to generate 3d plot.

        """
        from ..plotting import plot3d

        return plot3d(self, **kwargs)

    def plot2d(self, **kwargs):
        """Plot neuron in 2D using [`navis.plot2d`][].

        Parameters
        ----------
        **kwargs
                Keyword arguments will be passed to [`navis.plot2d`][].
                See `help(navis.plot2d)` for a list of accepted keywords.

        See Also
        --------
        [`navis.plot2d`][]
                Base function called to generate 2d plot.

        """
        from ..plotting import plot2d

        return plot2d(self, **kwargs)

    def summary(self,
                N: Optional[Union[int, slice]] = None,
                add_props: list = [],
                progress=False
                ) -> pd.DataFrame:
        """Get summary over all neurons in this NeuronList.

        Parameters
        ----------
        N :         int | slice, optional
                    If int, get only first N entries.
        add_props : list, optional
                    Additional properties to add to summary. If attribute not
                    available will return 'NA'.
        progress :  bool
                    Whether to show a progress bar. Can be useful for very
                    large list.

        Returns
        -------
        pandas DataFrame

        """
        if not self.empty:
            # Fetch a union of all summary props (keep order)
            all_props = [p for l in self.SUMMARY_PROPS for p in l]
            props = np.unique(all_props)
            props = sorted(props, key=lambda x: all_props.index(x))
        else:
            props = []

        # Add ID to properties - unless all are generic UUIDs
        if any([not isinstance(n.id, uuid.UUID) for n in self.neurons]):
            # Make sure we don't have two IDs
            if 'id' in props:
                props.remove('id')
            props = np.insert(props, 2, 'id')

        if add_props:
            props = np.append(props, add_props)

        if not isinstance(N, slice):
            N = slice(N)

        return pd.DataFrame(data=[[getattr(n, a, 'NA') for a in props]
                                  for n in config.tqdm(self.neurons[N],
                                                       desc='Summarizing',
                                                       leave=False,
                                                       disable=not progress)],
                            columns=props)

    def itertuples(self):
        """Helper to mimic `pandas.DataFrame.itertuples()`."""
        return self.neurons

    def add_metadata(self, meta, id_col='id', neuron_id='id', columns=None, register=False, missing='raise'):
        """Add neuron metadata from a DataFrame.

        Parameters
        ----------
        meta :      pd.DataFrame | str | Path
                    DataFrame or filepath to a CSV file containing metadata.
                    Must contain a column with neuron IDs.
        id_col :    str
                    Name of the column containing neuron IDs.
        neuron_id : str
                    Name of the attribute in the neuron that corresponds to
                    the `id_col`.
        columns :   list, optional
                    List of columns to add. If None, will add all columns except
                    for `id_col`.
        register :  bool
                    If True, will also register the attribute(s) as properties
                    that should show up in the summary.
        missing :   'raise' | 'warn' | 'ignore'
                    What to do if `meta` is missing a value for a neuron.

        See Also
        --------
        navis.NeuronList.set_neuron_attributes
                    Set individual attributes of neurons contained in the NeuronList.

        """
        assert missing in ('raise', 'warn', 'ignore')

        if isinstance(meta, (str, Path)):
            meta = pd.read_csv(meta)

        if not isinstance(meta, pd.DataFrame):
            raise TypeError('`meta` must be a DataFrame or a path to a CSV file, '
                            f'got {type(meta)}')

        if id_col not in meta.columns:
            raise KeyError(f'Column "{id_col}" not found in metadata.')

        # Index meta data by the neuron_id
        neuron_id = getattr(self, neuron_id)
        miss = ~np.isin(neuron_id, meta[id_col].values)
        if any(miss):
            msg = f'Metadata is missing entries for IDs: {neuron_id[miss]}'
            if missing == 'raise':
                raise KeyError(msg)
            elif missing == 'warn':
                logger.warning(msg)

        meta = meta.set_index(id_col).reindex(neuron_id)

        if columns is None:
            columns = meta.columns

        for c in columns:
            if c == id_col:
                continue
            self.set_neuron_attributes(
                meta[c].values.tolist(),
                name=c,
                register=register
                )

    def get_neuron_attributes(self, *args, **kwargs):
        """Get attributes of neurons contained in the NeuronList.

        Parameters
        ----------
        name :      str
                    Name of the property to get.
        default :   any, optional
                    Default value to return if attribute is not found.

        Returns
        -------
        np.ndarray
                    Array of values for the requested attribute.

        """
        return np.array([getattr(n, *args, **kwargs) for n in self.neurons])

    def set_neuron_attributes(self, x, name, register=False, na='raise'):
        """Set attributes of neurons contained in the NeuronList.

        Parameters
        ----------
        x :         any | list | np.ndarray | dict | function
                    Value of the property:
                      - lists and arrays are expected to contain a value for
                        each neuron and hence have to match the length of the
                        NeuronList
                      - dict is expected to map `{neuron.id: value}`
                      - a function is expected to take `neuron.id` as input
                        and return a value
        name :      str
                    Name of the property to set.
        register :  bool
                    If True, will also register the attribute(s) as properties
                    that should show up in the summary.
        na :        'raise' | 'propagate' | 'skip'
                    What to do if `x` is a dictionary and does not contain a
                    value for a neuron:
                     - 'raise' will raise a KeyError
                     - 'propagate' will set the attribute to `None`
                     - 'skip' will not set the attribute

        See Also
        --------
        navis.NeuronList.add_metadata
                    Set metadata from a dataframe.

        Examples
        --------
        >>> import navis
        >>> nl = navis.example_neurons(5)
        >>> # Set a single value
        >>> nl.set_neuron_attributes('some_value', name='my_attr')
        >>> nl[0].my_attr
        'some_value'
        >>> # Set individual values as iterable
        >>> nl.set_neuron_attributes([1, 2, 3, 4, 5], name='my_attr')
        >>> nl[0].my_attr
        1
        >>> nl.my_attr
        array([1, 2, 3, 4, 5])
        >>> # Set individual values using a dictionary
        >>> val_dict = dict(zip(nl.id, ['test', 2, 2.2, 4, 'test2']))
        >>> nl.set_neuron_attributes(val_dict, name='my_attr')
        >>> nl[0].my_attr
        'test'

        """
        utils.eval_param(na, name='na',
                         allowed_values=('raise', 'propagate', 'skip'))
        utils.eval_param(name, name='name', allowed_types=(str, ))

        if isinstance(x, dict):
            if na == 'raise':
                miss = ~np.isin(self.id, list(x))
                if any(miss):
                    raise KeyError('Dictionary `x` is missing entries for IDs: '
                                   f'{self.id[miss]}')
            for n in self.neurons:
                v = x.get(n.id, None)
                if (v is None) and (na == 'skip'):
                    continue
                n._register_attr(name, v, summary=register)
        elif isinstance(x, (list, np.ndarray)):
            if len(x) != len(self):
                raise ValueError(f'Got {len(x)} values for the{len(self)} '
                                 'neurons in the NeuronList.')
            for n, v in zip(self.neurons, x):
                n._register_attr(name, v, summary=register)
        elif callable(x):
            for n in self.neurons:
                n._register_attr(name, x(n.id), summary=register)
        else:
            for n in self.neurons:
                n._register_attr(name, x, summary=register)

    def sort_values(self, key: str, ascending: bool = False):
        """Sort neurons by given key.

        Needs to be an attribute of all neurons: for example `name`.
        Also works with custom attributes.
        """
        self.neurons = sorted(self.neurons,
                              key=lambda x: getattr(x, key),
                              reverse=ascending is False)

    def copy(self, **kwargs) -> 'NeuronList':
        """Return copy of this NeuronList.

        Parameters
        ----------
        **kwargs
                    Keyword arguments passed to neuron's `.copy()` method::

                    deepcopy :  bool, for TreeNeurons only
                                If False, `.graph` (NetworkX DiGraphs) will be
                                returned as views - changes to nodes/edges can
                                progagate back! `.igraph` (iGraph) - if
                                available - will always be deepcopied.

        """
        return self.__class__([n.copy(**kwargs) for n in config.tqdm(self.neurons,
                                                                     desc='Copy',
                                                                     leave=False,
                                                                     disable=config.pbar_hide or len(self) < 20)],
                              make_copy=False)

    def head(self, N: int = 5) -> pd.DataFrame:
        """Return summary for top N neurons."""
        return self.summary(N=N)

    def tail(self, N: int = 5) -> pd.DataFrame:
        """Return summary for bottom N neurons."""
        return self.summary(N=slice(-N, len(self)))

    def remove_duplicates(self,
                          key: str = 'name',
                          keep: str = 'first',
                          inplace: bool = False
                          ) -> Optional['NeuronList']:
        """Remove duplicate neurons from list.

        Parameters
        ----------
        key :       str | list, optional
                    Attribute(s) by which to identify duplicates. In case of
                    multiple, all attributes must match to flag a neuron as
                    duplicate.
        keep :      str
                    Which of the duplicated neurons to keep.
        inplace :   bool, optional
                    If False will return a copy of the original with
                    duplicates removed.

        """
        if inplace:
            x = self
        else:
            x = self.copy()

        key = utils.make_iterable(key)

        # Generate pandas DataFrame
        df = pd.DataFrame([[getattr(n, at) for at in key] for n in x],
                          columns=key)

        # Find out which neurons to keep
        to_keep = ~df.duplicated(keep=keep).values

        # Reassign neurons
        x.neurons = x[to_keep].neurons

        if not inplace:
            return x
        return None

    def unmix(self):
        """Split into NeuronLists of the same neuron type.

        Returns
        -------
        dict
                Dictionary of `{Neurontype: NeuronList}`

        """
        return {t: self.__class__([n for n in self.neurons if isinstance(n, t)])
                for t in self.types}


class _IdIndexer():
    """ID-based indexer for NeuronLists to access their neurons by ID."""

    def __init__(self, neuronlist):
        self.nl = neuronlist

    def __getitem__(self, ids):
        # Track if a single neuron was requested
        if not utils.is_iterable(ids):
            single = True
        else:
            single = False

        # Turn into list and force strings
        ids = utils.make_iterable(ids, force_type=str)

        # Generate a map
        # Note we account for the fact we might have duplicate IDs in the list
        map = {}
        for n in self.nl:
            map[str(n.id)] = map.get(str(n.id), []) + [n]

        # Get selection
        sel = [map.get(i, []) for i in ids]

        # Check for missing IDs
        miss = [i for i, k in zip(ids, sel) if len(k) == 0]
        if miss:
            raise ValueError(f'No neuron(s) found for ID(s): {", ".join(miss)}')

        # Check for duplicate Ids in query IDs or in resulting selection
        dupl = [i for i, k in zip(ids, sel) if len(k) > 1]
        if dupl or len(set(ids)) < len(ids):
            logger.warning('Selection contains duplicate IDs.')

        # Flatten selection
        sel = [n for l in sel for n in l]

        if single and len(sel) == 1:
            return sel[0]
        else:
            return self.nl.__class__(sel)



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import pandas as pd
import numpy as np


from abc import ABC, abstractmethod
from scipy.stats import wasserstein_distance
from typing import Union, Sequence

from .. import config, graph, core
from . import subset_neuron, tortuosity

# Set up logging
logger = config.get_logger(__name__)

__all__ = sorted(
    [
        "ivscc",
    ]
)

# A mapping of label IDs to compartment names
# Note: anything above 5 is considered "undefined" or "custom"
label_to_comp = {
    -1: "root",
    0: "undefined",
    1: "soma",
    2: "axon",
    3: "basal_dendrite",
    4: "apical_dendrite",
}
comp_to_label = {v: k for k, v in label_to_comp.items()}


class CompartmentNotFoundError(Exception):
    """An exception raised when a compartment is not found."""

    pass


class Features(ABC):
    def __init__(self, neuron: "core.TreeNeuron", label=None, verbose=False):
        self.neuron = neuron
        self.verbose = verbose

        if label is None:
            self.label = ""
        elif not label.endswith("_"):
            self.label = f"{label}_"
        else:
            self.label = label

        # Make sure the neuron is rooted to the soma (if present)
        self.soma = self.neuron.soma
        if self.soma is not None:
            self.soma_pos = self.neuron.soma_pos[0]
            self.soma_radius = self.neuron.nodes.set_index("node_id").loc[
                self.soma, "radius"
            ]

            if self.neuron.soma not in self.neuron.root:
                self.neuron = self.neuron.reroot(self.neuron.soma)

            # Calculate geodesic distances from leafs to all other nodes (directed)
            self.leaf_dists = graph.geodesic_matrix(
                self.neuron, self.neuron.leafs.node_id.values, directed=True
            )
            # Replace infinities with -1
            self.leaf_dists[self.leaf_dists == float("inf")] = -1

        self.features = {}

    def record_feature(self, name, value):
        """Record a feature."""
        self.features[f"{self.label}{name}"] = value

    @abstractmethod
    def extract_features(self):
        """Extract features."""
        pass


class BasicFeatures(Features):
    """Base class for features."""

    def extract_features(self):
        """Extract basic features."""
        self.record_feature(
            "extent_y", self.neuron.nodes.y.max() - self.neuron.nodes.y.min()
        )
        self.record_feature(
            "extent_x", self.neuron.nodes.x.max() - self.neuron.nodes.x.min()
        )
        self.record_feature(
            "max_branch_order", (self.neuron.nodes.type == "branch").sum() + 1
        )
        self.record_feature("num_nodes", len(self.neuron.nodes))
        self.record_feature("total_length", self.neuron.cable_length)

        if self.soma is None:
            if self.verbose:
                logger.warning(
                    f"{self.neuron.id} has no `.soma` attribute, skipping soma-related features."
                )
            return

        # x/y bias from soma
        # Note: this is absolute for x and relative for y
        self.record_feature(
            "bias_x",
            abs(
                (self.neuron.nodes.x.max() - self.soma_pos[0])
                - (self.soma_pos[0] - self.neuron.nodes.x.min())
            ),
        )
        self.record_feature(
            "bias_y",
            (self.neuron.nodes.y.max() - self.soma_pos[1])
            - (self.soma_pos[1] - self.neuron.nodes.y.min()),
        )

        # Distances from soma
        self.record_feature(
            "max_euclidean_distance",
            (
                (self.neuron.nodes[["x", "y", "z"]] - self.soma_pos)
                .pow(2)
                .sum(axis=1)
                .pow(0.5)
                .sum()
                .max()
            ),
        )
        self.record_feature(
            "max_path_length",
            self.leaf_dists.loc[
                self.leaf_dists.index.isin(self.neuron.nodes.node_id)
            ].values.max(),
        )

        # Tortuosity
        self.record_feature("mean_contraction", tortuosity(self.neuron))

        # Branching (number of linear segments between branch)
        self.record_feature("num_branches", len(self.neuron.small_segments))

        return self.features


class CompartmentFeatures(BasicFeatures):
    """Base class for compartment-specific features."""

    def __init__(self, neuron: "core.TreeNeuron", compartment, verbose=False):
        if "label" not in neuron.nodes.columns:
            raise ValueError(
                f"No 'label' column found in node table for neuron {neuron.id}"
            )

        if (
            compartment not in neuron.nodes.label.values
            and comp_to_label.get(compartment, compartment)
            not in neuron.nodes.label.values
        ):
            raise CompartmentNotFoundError(
                f"No {compartment} ({comp_to_label.get(compartment, compartment)}) compartments found in neuron {neuron.id}"
            )

        # Initialize the parent class
        super().__init__(neuron, label=compartment, verbose=verbose)

        # Now subset the neuron to this compartment
        self.neuron = subset_neuron(
            self.neuron,
            (
                self.neuron.nodes.label.isin(
                    (compartment, comp_to_label[compartment])
                ).values
            ),
        )


class AxonFeatures(CompartmentFeatures):
    """Extract features from an axon."""

    def __init__(self, neuron: "core.TreeNeuron", verbose=False):
        super().__init__(neuron, "axon", verbose=verbose)

    def extract_features(self):
        # Extract basic features via the parent class
        super().extract_features()

        # Now deal witha axon-specific features:

        if self.soma is not None:
            # Distance between axon root and soma surface
            # Note: we're catering for potentially multiple roots here
            axon_root_pos = self.neuron.nodes.loc[
                self.neuron.nodes.type == "root", ["x", "y", "z"]
            ].values

            # Closest dist between an axon root and the soma
            dist = np.linalg.norm(axon_root_pos - self.soma_pos, axis=1).min()

            # Subtract soma radius from the distance
            dist -= self.soma_radius

            self.record_feature("exit_distance", dist)

            # Axon theta: The relative radial position of the point where the neurite from which
            # the axon derives exits the soma.

            # Get the node where the axon exits the soma
            exit_node = self.neuron.nodes.loc[self.neuron.nodes.type == "root"]

            # Get theta
            theta = np.arctan2(
                exit_node.y.values - self.soma_pos[1],
                exit_node.x.values - self.soma_pos[0],
            )[0]
            self.record_feature("exit_theta", theta)

        return self.features


class BasalDendriteFeatures(CompartmentFeatures):
    """Extract features from a basal dendrite."""

    def __init__(self, neuron: "core.TreeNeuron", verbose=False):
        super().__init__(neuron, "basal_dendrite", verbose=verbose)

    def extract_features(self):
        # Extract basic features via the parent class
        super().extract_features()

        # Now deal with basal dendrite-specific features
        if self.soma is not None:
            # Number of stems sprouting from the soma
            # (i.e. number of nodes with a parent that is the soma)
            self.record_feature(
                "calculate_number_of_stems",
                (self.neuron.nodes.parent_id == self.soma).sum(),
            )

        return self.features


class ApicalDendriteFeatures(CompartmentFeatures):
    """Extract features from a apical dendrite."""

    def __init__(self, neuron: "core.TreeNeuron", verbose=False):
        super().__init__(neuron, "apical_dendrite", verbose=verbose)

    def extract_features(self):
        # Extract basic features via the parent class
        super().extract_features()

        return self.features


class OverlapFeatures(Features):
    """Features that compare two compartments (e.g. overlap)."""

    # Compartments to compare
    compartments = ("axon", "basal_dendrite", "apical_dendrite")

    def extract_features(self):
        # Iterate over compartments
        for c1 in self.compartments:
            if c1 in self.neuron.nodes.label.values:
                c1_nodes = self.neuron.nodes[self.neuron.nodes.label == c1]
            elif comp_to_label.get(c1, c1) in self.neuron.nodes.label.values:
                c1_nodes = self.neuron.nodes[
                    self.neuron.nodes.label == comp_to_label[c1]
                ]
            else:
                continue
            for c2 in self.compartments:
                if c1 == c2:
                    continue
                if c2 in self.neuron.nodes.label.values:
                    c2_nodes = self.neuron.nodes[self.neuron.nodes.label == c2]
                elif comp_to_label.get(c2, c2) in self.neuron.nodes.label.values:
                    c2_nodes = self.neuron.nodes[
                        self.neuron.nodes.label == comp_to_label[c2]
                    ]
                else:
                    continue

                # Calculate % of nodes of a given compartment type above/overlapping/below the
                # full y-extent of another compartment type
                self.features[f"{c1}_frac_above_{c2}"] = (
                    c1_nodes.y > c2_nodes.y.max()
                ).sum() / len(c1_nodes)
                self.features[f"{c1}_frac_intersect_{c2}"] = (
                    (c1_nodes.y >= c2_nodes.y.min()) & (c1_nodes.y <= c2_nodes.y.max())
                ).sum() / len(c1_nodes)
                self.features[f"{c1}_frac_below_{c2}"] = (
                    c1_nodes.y < c2_nodes.y.min()
                ).sum() / len(c1_nodes)

                # Calculate earth mover's distance (EMD) between the two compartments
                if f"{c2}_emd_with_{c1}" not in self.features:
                    self.features[f"{c1}_emd_with_{c2}"] = wasserstein_distance(
                        c1_nodes.y, c2_nodes.y
                    )

        return self.features


def ivscc_features(
    x: "core.TreeNeuron",
    features=None,
    missing_compartments="ignore",
    verbose=False,
    progress=True,
) -> Union[float, pd.DataFrame]:
    """Calculate IVSCC features for neuron(s).

    Please see the `IVSCC` tutorial for details.

    Parameters
    ----------
    x :                     TreeNeuron | NeuronList
                            Neuron(s) to calculate IVSCC for.
    features :              Sequence[Features], optional
                            Provide specific features to calculate.
                            Must be subclasses of `BasicFeatures`.
                            If `None`, will use default features.
    missing_compartments : "ignore" | "skip" | "raise"
                            What to do if a neuron is missing a compartment
                            (e.g. no axon or basal dendrite):
                             - "ignore" (default): ignore that compartment
                             - "skip": skip the entire neuron
                             - "raise": raise an exception

    Returns
    -------
    ivscc :                 pd.DataFrame
                            IVSCC features for the neuron(s).

    """

    if isinstance(x, core.TreeNeuron):
        x = core.NeuronList([x])

    if features is None:
        features = DEFAULT_FEATURES

    data = {}
    for n in config.tqdm(
        x, desc="Calculating IVSCC features", disable=not progress or config.pbar_hide
    ):
        data[n.id] = {}
        for feat in features:
            try:
                f = feat(n, verbose=verbose)
            except CompartmentNotFoundError as e:
                if missing_compartments == "ignore":
                    continue
                elif missing_compartments == "skip":
                    if verbose:
                        print(f"Skipping neuron {n.id}: {e}")
                    data.pop(n.id)
                    break
                else:
                    raise e

            data[n.id].update(f.extract_features())

    return pd.DataFrame(data)


def _check_compartments(n, compartments):
    """Check if `compartments` are valid."""
    if compartments == "auto":
        if "label" not in n.nodes.columns:
            return None
        return n.nodes.label.unique()
    elif compartments is True:
        return n.nodes.label.unique()
    elif isinstance(compartments, str):
        if "label" not in n.nodes.columns or compartments not in n.nodes.label.unique():
            raise ValueError(f"Compartment not present: {compartments}")
        return [compartments]
    elif isinstance(compartments, Sequence):
        if "label" not in n.nodes.columns:
            raise ValueError("No 'label' column found in node table.")
        for c in compartments:
            if c not in n.nodes.label.unique():
                raise ValueError(f"Compartment not present: {c}")
        return compartments
    elif compartments in (None, False):
        return None

    raise ValueError(f"Invalid `compartments`: {compartments}")


DEFAULT_FEATURES = [
    AxonFeatures,
    BasalDendriteFeatures,
    ApicalDendriteFeatures,
    OverlapFeatures,
]



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2017 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along

"""Functions to generate vispy visuals."""

import uuid
import warnings

import pandas as pd
import numpy as np

import matplotlib.colors as mcl

from ... import core, config, utils, conversion
from ..colors import prepare_colormap, vertex_colors, eval_color

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import vispy
    from vispy import scene
    from vispy.geometry import create_sphere

__all__ = [
    "volume2vispy",
    "neuron2vispy",
    "dotprop2vispy",
    "voxel2vispy",
    "points2vispy",
    "combine_visuals",
]

logger = config.get_logger(__name__)


def volume2vispy(x, settings):
    """Convert Volume(s) to vispy visuals."""
    # Must not use make_iterable here as this will turn into list of keys!
    if not isinstance(x, (list, np.ndarray)):
        x = [x]

    # List to fill with vispy visuals
    visuals = []
    for i, v in enumerate(x):
        if not isinstance(v, core.Volume):
            raise TypeError(f'Expected navis.Volume, got "{type(v)}"')

        object_id = uuid.uuid4()

        if settings.color is not None:
            color = settings.color
        else:
            color = getattr(v, "color", (0.95, 0.95, 0.95, 0.1))

        # Colors might be list, need to pick the correct color for this volume
        if isinstance(color, list):
            if all([isinstance(c, (tuple, list, np.ndarray)) for c in color]):
                color = color[i]

        if isinstance(color, str):
            color = mcl.to_rgb(color)

        color = np.array(color, dtype=float)

        # Add alpha
        if len(color) < 4:
            color = np.append(color, [0.1])

        if max(color) > 1:
            color[:3] = color[:3] / 255

        s = scene.visuals.Mesh(
            vertices=v.vertices,
            faces=v.faces,
            color=color,
            shading=settings.shading,
        )

        # Set some aesthetic parameters
        # Note that for larger meshes adjusting the shading filter actually
        # surprisingly slow (e.g. ~4s @ 400k faces). Since volumes typically
        # don't have too many faces, we will keep setting the shininess.
        if int(vispy.__version__.split(".")[1]) >= 7:
            s.shading_filter.shininess = 0
        else:
            s.shininess = 0

        # Possible presets are "additive", "translucent", "opaque"
        s.set_gl_state("additive", cull_face=True, depth_test=False)
        # s.set_gl_state('additive' if color[3] < 1 else 'opaque',
        #               cull_face=True,
        #               depth_test=False if color[3] < 1 else True)

        # Make sure volumes are always drawn after neurons
        s.order = 10

        # Add custom attributes
        s.unfreeze()
        s._object_type = "volume"
        s._volume_name = getattr(v, "name", None)
        s._object = v
        s._object_id = object_id
        s.freeze()

        visuals.append(s)

    return visuals


def neuron2vispy(x, settings):
    """Convert a Neuron/List to vispy visuals.

    Parameters
    ----------
    x :               TreeNeuron | MeshNeuron | Dotprops | VoxelNeuron | NeuronList
                      Neuron(s) to plot.
    color :           list | tuple | array | str
                      Color to use for plotting.
    colormap :        tuple | dict | array
                      Color to use for plotting. Dictionaries should be mapped
                      by ID. Overrides `color`.
    connectors :      bool, optional
                      If True, plot connectors.
    connectors_only : bool, optional
                      If True, only connectors are plotted.
    by_strahler :     bool, optional
                      If True, shade neurites by strahler order.
    by_confidence :   bool, optional
                      If True, shade neurites by confidence.
    linewidth :       int, optional
                      Set linewidth. Might not work depending on your backend.
    cn_mesh_colors :  bool, optional
                      If True, connectors will have same color as the neuron.
    cn_layout :       dict, optional
                      Sets connector (e.g. synapse) layout. For example::

                        {
                            0: {
                                'name': 'Presynapses',
                                'color': (255, 0, 0)
                                },
                            1: {
                                'name': 'Postsynapses',
                                'color': (0, 0, 255)
                                },
                            2: {
                                'name': 'Gap junctions',
                                'color': (0, 255, 0)
                                },
                            'display': 'lines'  # can also be 'circles'
                        }

    Returns
    -------
    list
                    Contains vispy visuals for each neuron.

    """
    if isinstance(x, core.BaseNeuron):
        x = core.NeuronList(x)
    elif isinstance(x, core.NeuronList):
        pass
    else:
        raise TypeError(f'Unable to process data of type "{type(x)}"')

    # colors = kwargs.get('color',
    #                     kwargs.get('c',
    #                                kwargs.get('colors', None)))
    # palette = kwargs.get('palette', None)
    # color_by = kwargs.get('color_by', None)
    # shade_by = kwargs.get('shade_by', None)

    # Color_by can be a per-node/vertex color, or a per-neuron color
    # such as property of the neuron
    color_neurons_by = None
    if settings.color_by is not None and x:
        # Check if this is a neuron property
        if isinstance(settings.color_by, str):
            if hasattr(x[0], settings.color_by):
                # If it is, use it to color neurons
                color_neurons_by = [getattr(neuron, settings.color_by) for neuron in x]
                settings.color_by = None
        elif isinstance(settings.color_by, (list, np.ndarray)):
            if len(settings.color_by) == len(x):
                color_neurons_by = settings.color_by
                settings.color_by = None

    if not isinstance(settings.color_by, type(None)):
        if not settings.palette:
            raise ValueError(
                'Must provide `palette` (e.g. "viridis") argument '
                "if using `color_by`"
            )

        colormap = vertex_colors(
            x,
            by=settings.color_by,
            alpha=settings.alpha,
            palette=settings.palette,
            vmin=settings.vmin,
            vmax=settings.vmax,
            na="raise",
            color_range=1,
        )
    else:
        colormap, _ = prepare_colormap(
            settings.color,
            neurons=x,
            palette=settings.palette,
            alpha=settings.alpha,
            color_by=color_neurons_by,
            color_range=1,
        )

    if not isinstance(settings.shade_by, type(None)):
        alphamap = vertex_colors(
            x,
            by=settings.shade_by,
            use_alpha=True,
            palette="viridis",  # palette is irrelevant here
            vmin=settings.smin,
            vmax=settings.smax,
            na="raise",
            color_range=1,
        )

        new_colormap = []
        for c, a in zip(colormap, alphamap):
            if not (isinstance(c, np.ndarray) and c.ndim == 2):
                c = np.tile(c, (a.shape[0], 1))

            if c.shape[1] == 4:
                c[:, 3] = a[:, 3]
            else:
                c = np.insert(c, 3, a[:, 3], axis=1)

            new_colormap.append(c)
        colormap = new_colormap

    # List to fill with vispy visuals
    visuals = []
    _radius_warned = False
    for i, neuron in enumerate(x):
        # Generate random ID -> we need this in case we have duplicate IDs
        object_id = uuid.uuid4()

        if isinstance(neuron, core.TreeNeuron) and settings.radius == "auto":
            # Number of nodes with radii
            n_radii = (neuron.nodes.get("radius", pd.Series([])).fillna(0) > 0).sum()
            # If less than 30% of nodes have a radius, we will fall back to lines
            if n_radii / neuron.nodes.shape[0] < 0.3:
                settings.radius = False

        if isinstance(neuron, core.TreeNeuron) and settings.radius:
            # Warn once if more than 5% of nodes have missing radii
            if not _radius_warned:
                if (
                    (neuron.nodes.radius.fillna(0).values <= 0).sum() / neuron.n_nodes
                ) > 0.05:
                    logger.warning(
                        "Some skeleton nodes have radius <= 0. This may lead to "
                        "rendering artifacts. Set `radius=False` to plot skeletons "
                        "as single-width lines instead."
                    )
                    _radius_warned = True

            _neuron = conversion.tree2meshneuron(
                neuron,
                warn_missing_radii=False,
                radius_scale_factor=settings.get("linewidth", 1),
            )
            _neuron.connectors = neuron.connectors
            neuron = _neuron

            # See if we need to map colors to vertices
            if isinstance(colormap[i], np.ndarray) and colormap[i].ndim == 2:
                colormap[i] = colormap[i][neuron.vertex_map]

        neuron_color = colormap[i]
        if not settings.connectors_only:
            if isinstance(neuron, core.TreeNeuron):
                visuals += skeleton2vispy(neuron, neuron_color, object_id, settings)
            elif isinstance(neuron, core.MeshNeuron):
                visuals += mesh2vispy(neuron, neuron_color, object_id, settings)
            elif isinstance(neuron, core.Dotprops):
                visuals += dotprop2vispy(neuron, neuron_color, object_id, settings)
            elif isinstance(neuron, core.VoxelNeuron):
                visuals += voxel2vispy(neuron, neuron_color, object_id, settings)
            else:
                logger.warning(
                    f"Don't know how to plot neuron of type '{type(neuron)}'"
                )

        if settings.connectors or settings.connectors_only and neuron.has_connectors:
            visuals += connectors2vispy(neuron, neuron_color, object_id, settings)

    return visuals


def connectors2vispy(neuron, neuron_color, object_id, settings):
    """Convert connectors to vispy visuals."""
    cn_lay = config.default_connector_colors.copy()
    cn_lay.update(settings.cn_layout)

    if isinstance(settings.connectors, (list, np.ndarray, tuple)):
        connectors = neuron.connectors[neuron.connectors.type.isin(settings.connectors)]
    elif settings.connectors == "pre":
        connectors = neuron.presynapses
    elif settings.connectors == "post":
        connectors = neuron.postsynapses
    elif isinstance(settings.connectors, str):
        connectors = neuron.connectors[neuron.connectors.type == settings.connectors]
    else:
        connectors = neuron.connectors

    visuals = []
    for j, this_cn in connectors.groupby("type"):
        if isinstance(settings.cn_colors, dict):
            color = settings.cn_colors.get(
                j, cn_lay.get(j, {}).get("color", (0.1, 0.1, 0.1))
            )
        elif settings.cn_colors == "neuron":
            color = neuron_color
        elif settings.cn_colors is not None:
            color = settings.cn_colors
        else:
            color = cn_lay.get(j, {}).get("color", (0.1, 0.1, 0.1))

        color = eval_color(color, color_range=1)

        pos = this_cn[["x", "y", "z"]].apply(pd.to_numeric).values

        mode = cn_lay["display"]
        if mode == "circles" or isinstance(neuron, core.MeshNeuron):
            con = scene.visuals.Markers(
                spherical=cn_lay.get("spherical", True),
                scaling=cn_lay.get("scale", False),
            )

            con.set_data(
                pos=np.array(pos),
                face_color=color,
                edge_color=color,
                size=settings.cn_size if settings.cn_size else cn_lay["size"],
            )

        elif mode == "lines":
            tn_coords = (
                neuron.nodes.set_index("node_id")
                .loc[this_cn.node_id.values][["x", "y", "z"]]
                .apply(pd.to_numeric)
                .values
            )

            segments = [item for sublist in zip(pos, tn_coords) for item in sublist]

            con = scene.visuals.Line(
                pos=np.array(segments),
                color=color,
                # Can only be used with method 'agg'
                width=settings.linewidth,
                connect="segments",
                antialias=False,
                method="gl",
            )
            # method can also be 'agg' -> has to use connect='strip'
        else:
            raise ValueError(f'Unknown connector display mode "{mode}"')

        # Add custom attributes
        con.unfreeze()
        con._object_type = "neuron"
        con._neuron_part = "connectors"
        con._id = neuron.id
        con._name = str(getattr(neuron, "name", neuron.id))
        con._object_id = object_id
        con._object = neuron
        con.freeze()

        visuals.append(con)
    return visuals


def mesh2vispy(neuron, neuron_color, object_id, settings):
    """Convert mesh (i.e. MeshNeuron) to vispy visuals."""
    # Skip empty neurons
    if not len(neuron.faces):
        return []

    color_kwargs = dict(color=neuron_color)
    if isinstance(neuron_color, np.ndarray) and neuron_color.ndim == 2:
        if len(neuron_color) == len(neuron.vertices):
            color_kwargs = dict(vertex_colors=neuron_color)
        elif len(neuron_color) == len(neuron.faces):
            color_kwargs = dict(face_colors=neuron_color)
        else:
            color_kwargs = dict(color=neuron_color)

    # There is a bug in pickling/unpickling numpy arrays where an internal flag
    # is set from 1 to 0 -> this then in turn upsets vispy. See also:
    # https://github.com/vispy/vispy/issues/1741#issuecomment-574425662
    # Pickling happens when meshneurons have been multi-processed. To fix this
    # the best way is to use astype (.copy() doesn't cut it)
    if neuron.vertices.dtype.isbuiltin == 0:
        neuron.vertices = neuron.vertices.astype(neuron.vertices.dtype.str)
    if neuron.faces.dtype.isbuiltin == 0:
        neuron.faces = neuron.faces.astype(neuron.faces.dtype.str)

    m = scene.visuals.Mesh(
        vertices=neuron.vertices,
        faces=neuron.faces,
        shading=settings.shading,
        **color_kwargs,
    )

    # Set some aesthetic parameters
    # Vispy 0.7.0 uses a new shading filter
    # Note that for larger meshes adjusting the shading filter is actually
    # surprisingly slow (e.g. ~4s @ 400k faces)
    if isinstance(settings.shininess, (int, float)):
        if int(vispy.__version__.split(".")[1]) >= 7:
            m.shading_filter.shininess = settings.shininess
        else:
            m.shininess = settings.shininess

    # Possible presets are "additive", "translucent", "opaque"
    if len(neuron_color) == 4 and neuron_color[3] < 1:
        m.set_gl_state("additive", cull_face=True, depth_test=False)

    # Add custom attributes
    m.unfreeze()
    m._object_type = "neuron"
    m._neuron_part = "neurites"
    m._id = neuron.id
    m._name = str(getattr(neuron, "name", neuron.id))
    m._object_id = object_id
    m._object = neuron
    m.freeze()
    return [m]


def to_vispy_cmap(color, fade=True):
    """Convert a given colour to a vispy colormap."""
    # First force RGB
    stop = mcl.to_rgba(color)
    start = mcl.to_rgba(color, alpha=0)

    # Convert to vispy cmap
    colors = vispy.color.ColorArray([start, stop])
    cmap = vispy.color.colormap.Colormap(colors=colors)

    # cmap consists of two colors
    # We will set first color to black and transparent
    if fade:
        col_arr = cmap.colors.rgba
        col_arr[1][:] = 0
        cmap.colors.rgba = col_arr

    return cmap


def voxel2vispy(neuron, neuron_color, object_id, settings):
    """Convert voxels (i.e. VoxelNeuron) to vispy visuals."""
    # Note the transpose: currently, vispy expects zyx for volumes but this
    # might change in the future
    grid = neuron.grid.T
    # Vispy doesn't like boolean matrices here
    if grid.dtype == bool:
        grid = grid.astype(int)
    vx = scene.visuals.Volume(vol=grid, cmap=to_vispy_cmap(neuron_color))

    # Add transforms
    vx.set_transform("st", scale=neuron.units_xyz.magnitude, translate=neuron.offset)

    # Add custom attributes
    vx.unfreeze()
    vx._object_type = "neuron"
    vx._neuron_part = "neurites"
    vx._id = neuron.id
    vx._name = str(getattr(neuron, "name", neuron.id))
    vx._object_id = object_id
    vx._object = neuron
    vx.freeze()
    return [vx]


def skeleton2vispy(neuron, neuron_color, object_id, settings):
    """Convert skeleton (i.e. TreeNeuron) into vispy visuals."""
    if neuron.nodes.empty:
        logger.warning(f"Skipping TreeNeuron w/o nodes: {neuron.id}")
        return []
    elif neuron.nodes.shape[0] == 1:
        logger.warning(f"Skipping single-node TreeNeuron: {neuron.label}")
        return []

    visuals = []
    if not settings.connectors_only:
        # Make sure we have one color for each node
        neuron_color = np.asarray(neuron_color)
        if neuron_color.ndim == 1:
            neuron_color = np.tile(neuron_color, (neuron.nodes.shape[0], 1))

        # Get nodes
        non_roots = neuron.nodes[neuron.nodes.parent_id >= 0]
        connect = np.zeros((non_roots.shape[0], 2), dtype=int)
        node_ix = pd.Series(
            np.arange(neuron.nodes.shape[0]), index=neuron.nodes.node_id.values
        )
        connect[:, 0] = node_ix.loc[non_roots.node_id].values
        connect[:, 1] = node_ix.loc[non_roots.parent_id].values

        # Create line plot from segments.
        t = scene.visuals.Line(
            pos=neuron.nodes[["x", "y", "z"]].values,
            color=neuron_color,
            # Can only be used with method 'agg'
            width=settings.linewidth,
            connect=connect,
            antialias=True,
            method="gl",
        )
        # method can also be 'agg' -> has to use connect='strip'
        # Make visual discoverable
        t.interactive = True

        # Add custom attributes
        t.unfreeze()
        t._object_type = "neuron"
        t._neuron_part = "neurites"
        t._id = neuron.id
        t._name = str(getattr(neuron, "name", neuron.id))
        t._object = neuron
        t._object_id = object_id
        t.freeze()

        visuals.append(t)

        # Extract and plot soma
        soma = utils.make_iterable(neuron.soma)
        if settings.soma:
            # If soma detection is messed up we might end up producing
            # hundrets of soma which will freeze the session
            if len(soma) >= 10:
                logger.warning(
                    f"Neuron {neuron.id} appears to have {len(soma)}"
                    " somas. That does not look right - will ignore "
                    "them for plotting."
                )
            else:
                for s in soma:
                    # Skip `None` somas
                    if isinstance(s, type(None)):
                        continue

                    # If we have colors for every vertex, we need to find the
                    # color that corresponds to this root (or it's parent to be
                    # precise)
                    if isinstance(neuron_color, np.ndarray) and neuron_color.ndim > 1:
                        s_ix = np.where(neuron.nodes.node_id == s)[0][0]
                        soma_color = neuron_color[s_ix]
                    else:
                        soma_color = neuron_color

                    n = neuron.nodes.set_index("node_id").loc[s]
                    r = (
                        getattr(n, neuron.soma_radius)
                        if isinstance(neuron.soma_radius, str)
                        else neuron.soma_radius
                    )
                    sp = create_sphere(7, 7, radius=r)
                    verts = sp.get_vertices() + n[["x", "y", "z"]].values.astype(
                        np.float32
                    )
                    s = scene.visuals.Mesh(
                        vertices=verts,
                        shading="smooth",
                        faces=sp.get_faces(),
                        color=soma_color,
                    )
                    # Vispy 0.7.0 uses a new shading filter
                    if int(vispy.__version__.split(".")[1]) >= 7:
                        s.shading_filter.ambient_light = vispy.color.Color("white")
                        s.shading_filter.shininess = 0
                    else:
                        s.ambient_light_color = vispy.color.Color("white")
                        s.shininess = 0

                    # Make visual discoverable
                    s.interactive = True

                    # Add custom attributes
                    s.unfreeze()
                    s._object_type = "neuron"
                    s._neuron_part = "soma"
                    s._id = neuron.id
                    s._name = str(getattr(neuron, "name", neuron.id))
                    s._object = neuron
                    s._object_id = object_id
                    s.freeze()

                    visuals.append(s)

    return visuals


def dotprop2vispy(x, neuron_color, object_id, settings):
    """Convert dotprops(s) to vispy visuals.

    Parameters
    ----------
    x :             navis.Dotprops | pd.DataFrame
                    Dotprop(s) to plot.

    Returns
    -------
    list
                    Contains vispy visuals for each dotprop.

    """
    # Skip empty neurons
    if not len(x.points):
        return []

    # Generate TreeNeuron
    tn = x.to_skeleton(scale_vec=settings.dps_scale_vec)
    return skeleton2vispy(tn, neuron_color, object_id, settings)


def points2vispy(x, **kwargs):
    """Convert points to vispy visuals.

    Parameters
    ----------
    x :             list of arrays
                    Points to plot.
    color :         tuple | array
                    Color to use for plotting.
    size :          int, optional
                    Marker size.

    Returns
    -------
    list
                    Contains vispy visuals for points.

    """
    colors = kwargs.get(
        "color",
        kwargs.get("c", kwargs.get("colors", eval_color(config.default_color, 1))),
    )

    visuals = []
    for p in x:
        object_id = uuid.uuid4()
        if not isinstance(p, np.ndarray):
            p = np.array(p)

        con = scene.visuals.Markers(
            spherical=kwargs.get("spherical", True), scaling=kwargs.get("scale", False)
        )
        con.set_data(
            pos=p,
            face_color=colors,
            edge_color=colors,
            size=kwargs.get("size", kwargs.get("s", 2)),
        )

        # Add custom attributes
        con.unfreeze()
        con._object_type = "points"
        con._object_id = object_id
        con.freeze()

        visuals.append(con)

    return visuals


def combine_visuals(visuals, name=None):
    """Attempt to combine multiple visuals of similar type into one.

    Parameters
    ----------
    visuals :   List
                List of visuals.
    name :      str, optional
                Legend name for the combined visual.

    Returns
    -------
    list
                List of visuals some of which where combined.

    """
    if any([not isinstance(v, scene.visuals.VisualNode) for v in visuals]):
        raise TypeError("Visuals must all be instances of VisualNode")

    # Combining visuals (i.e. adding pos AND changing colors) fails if
    # they are already on a canvas
    if any([v.parent for v in visuals]):
        raise ValueError("Visuals must not have parents when combined.")

    # Sort into types
    types = set([type(v) for v in visuals])

    by_type = {ty: [v for v in visuals if type(v) is ty] for ty in types}

    combined = []

    # Now go over types and combine when possible
    for ty in types:
        # Skip if nothing to combine
        if len(by_type[ty]) <= 1:
            combined += by_type[ty]
            continue

        if ty == scene.visuals.Line:
            # Collate data
            pos = np.concatenate([vis._pos for vis in by_type[ty]])

            # We need to produce one color/vertex and offset the connections
            colors = []
            connect = []
            offset = 0
            for vis in by_type[ty]:
                if vis.color.ndim == 2:
                    colors.append(vis.color)
                else:
                    colors.append(np.repeat([vis.color], vis.pos.shape[0], axis=0))
                connect.append(vis._connect + offset)
                offset += vis._pos.shape[0]

            connect = np.concatenate(connect)
            colors = np.concatenate(colors)

            t = scene.visuals.Line(
                pos=pos,
                color=colors,
                # Can only be used with method 'agg'
                connect=connect,
                antialias=True,
                method="gl",
            )
            # method can also be 'agg' -> has to use connect='strip'
            # Make visual discoverable
            t.interactive = True

            # Add custom attributes
            t.unfreeze()
            t._object_type = "neuron"
            t._neuron_part = "neurites"
            t._id = "NA"
            t._object_id = uuid.uuid4()
            t._name = name if name else "NeuronCollection"
            t.freeze()

            combined.append(t)
        elif ty == scene.visuals.Mesh:
            vertices = []
            faces = []
            color = []
            for vis in by_type[ty]:
                verts_offset = sum([v.shape[0] for v in vertices])
                faces.append(vis.mesh_data.get_faces() + verts_offset)
                vertices.append(vis.mesh_data.get_vertices())

                vc = vis.mesh_data.get_vertex_colors()
                if not isinstance(vc, type(None)):
                    color.append(vc)
                else:
                    color.append(
                        np.tile(vis.color.rgba, len(vertices[-1])).reshape(-1, 4)
                    )

            faces = np.vstack(faces)
            vertices = np.vstack(vertices)
            color = np.concatenate(color)

            if np.unique(color, axis=0).shape[0] == 1:
                base_color = color[0]
            else:
                base_color = (1, 1, 1, 1)

            t = scene.visuals.Mesh(
                vertices,
                faces=faces,
                color=base_color,
                vertex_colors=color,
                shading=by_type[ty][0].shading,
                mode=by_type[ty][0].mode,
            )

            # Add custom attributes
            t.unfreeze()
            t._object_type = "neuron"
            t._neuron_part = "neurites"
            t._id = "NA"
            t._name = name if name else "MeshNeuronCollection"
            t._object_id = uuid.uuid4()
            t.freeze()

            combined.append(t)
        else:
            combined += by_type[ty]

    return combined



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.


import pandas as pd
import numpy as np
import networkx as nx

from typing import Union, Sequence, Callable

from .. import utils, config, core, graph

# Set up logging
logger = config.get_logger(__name__)

__all__ = sorted(["subset_neuron"])


@utils.map_neuronlist(desc="Subsetting", allow_parallel=True)
@utils.lock_neuron
def subset_neuron(
    x: Union["core.TreeNeuron", "core.MeshNeuron"],
    subset: Union[Sequence[Union[int, str]], nx.DiGraph, pd.DataFrame, Callable],
    inplace: bool = False,
    keep_disc_cn: bool = False,
    prevent_fragments: bool = False,
) -> "core.NeuronObject":
    """Subset a neuron to a given set of nodes/vertices.

    Note that for `MeshNeurons` it is not guaranteed that all vertices in
    `subset` survive because we will also drop degenerate vertices that do
    not participate in any faces.

    Parameters
    ----------
    x :                   TreeNeuron | MeshNeuron | Dotprops | NeuronList
                          Neuron to subset. When passing a NeuronList, it's advised
                          to use a function for `subset` (see below).
    subset :              list-like | set | NetworkX.Graph | pandas.DataFrame | Callable
                          Subset of the neuron to keep. Depending on the neuron:
                            For TreeNeurons:
                             - node IDs
                             - a boolean mask matching the number of nodes
                             - DataFrame with `node_id` column
                            For MeshNeurons:
                             - vertex indices
                             - a boolean mask matching either the number of
                               vertices or faces
                            For Dotprops:
                             - point indices
                             - a boolean mask matching the number of points
                          Alternatively, you can pass a function that accepts
                          a neuron and returns a suitable `subset` as described
                          above. This is useful e.g. when wanting to subset a
                          list of neurons.
    keep_disc_cn :        bool, optional
                          If False, will remove disconnected connectors that
                          have "lost" their parent node/vertex.
    prevent_fragments :   bool, optional
                          If True, will add nodes/vertices to `subset`
                          required to keep neuron from fragmenting. Ignored for
                          `Dotprops`.
    inplace :             bool, optional
                          If False, a copy of the neuron is returned.

    Returns
    -------
    TreeNeuron | MeshNeuron | Dotprops | NeuronList

    Examples
    --------
    Subset skeleton to all branches with less than 10 nodes

    >>> import navis
    >>> # Get neuron
    >>> n = navis.example_neurons(1)
    >>> # Get all linear segments
    >>> segs = n.segments
    >>> # Get short segments
    >>> short_segs = [s for s in segs if len(s) <= 10]
    >>> # Flatten segments into list of nodes
    >>> nodes_to_keep = [n for s in short_segs for n in s]
    >>> # Subset neuron
    >>> n_short = navis.subset_neuron(n, subset=nodes_to_keep)

    Subset multiple neurons using a callable

    >>> import navis
    >>> nl = navis.example_neurons(2)
    >>> # Subset neurons to all leaf nodes
    >>> nl_end = navis.subset_neuron(
    ...     nl,
    ...     subset=lambda x: x.leafs.node_id
    ... )

    See Also
    --------
    [`navis.cut_skeleton`][]
            Cut neuron at specific points.
    [`navis.in_volume`][]
            To intersect a neuron with a volume (mesh).

    """
    if isinstance(x, core.NeuronList) and len(x) == 1:
        x = x[0]

    utils.eval_param(
        x, name="x", allowed_types=(core.TreeNeuron, core.MeshNeuron, core.Dotprops)
    )

    if callable(subset):
        subset = subset(x)

    # Make a copy of the neuron
    if not inplace:
        x = x.copy()
        # We have to run this in a separate function so that the lock is applied
        # to the copy
        subset_neuron(
            x,
            subset=subset,
            inplace=True,
            keep_disc_cn=keep_disc_cn,
            prevent_fragments=prevent_fragments,
        )
        return x

    if isinstance(x, core.TreeNeuron):
        x = _subset_treeneuron(
            x,
            subset=subset,
            keep_disc_cn=keep_disc_cn,
            prevent_fragments=prevent_fragments,
        )
    elif isinstance(x, core.MeshNeuron):
        x = _subset_meshneuron(
            x,
            subset=subset,
            keep_disc_cn=keep_disc_cn,
            prevent_fragments=prevent_fragments,
        )
    elif isinstance(x, core.Dotprops):
        x = _subset_dotprops(x, subset=subset, keep_disc_cn=keep_disc_cn)

    return x


def _subset_dotprops(x, subset, keep_disc_cn):
    """Subset Dotprops."""
    if not utils.is_iterable(subset):
        raise TypeError(
            "Can only subset Dotprops to list, set or "
            f'numpy.ndarray, not "{type(subset)}"'
        )

    subset = utils.make_iterable(subset)

    # Convert indices to mask
    if subset.dtype == bool:
        if subset.shape != (x.points.shape[0],):
            raise ValueError("Boolean mask must be of same length as points.")
        mask = subset
    else:
        mask = np.isin(np.arange(0, len(x.points)), subset)

    # Filter connectors
    if not keep_disc_cn and x.has_connectors:
        if "point" not in x.connectors.columns:
            x.connectors["point"] = x.snap(x.connectors[["x", "y", "z"]].values)[0]

        if subset.dtype == bool:
            subset = np.arange(0, len(x.points))[subset]

        x._connectors = x.connectors[x.connectors.point.isin(subset)].copy()
        x._connectors.reset_index(inplace=True, drop=True)

        # Make old -> new indices map
        new_ix = dict(zip(subset, np.arange(0, len(subset))))

        x.connectors["point"] = x.connectors.point.map(new_ix)

    # Mask vectors
    # This will also trigger re-calculation which is necessary for two reasons:
    # 1. Vectors will change if they have to be recalculated from
    #    the downsampled dotprops.
    # 2. There might not be enough points left after downsampling given the
    #    original k.
    if isinstance(x._vect, type(None)) and x.k:
        if x.n_points >= x.k:
            x.recalculate_tangents(k=x.k, inplace=True)
    x._vect = x._vect[mask]

    # Mask alphas if exists
    if not isinstance(x._alpha, type(None)):
        x._alpha = x._alpha[mask]

    # Finally mask points
    x._points = x._points[mask]

    return x


def _subset_meshneuron(x, subset, keep_disc_cn, prevent_fragments):
    """Subset MeshNeuron."""
    if not utils.is_iterable(subset):
        raise TypeError(
            "Can only subset MeshNeuron to list, set or "
            f'numpy.ndarray, not "{type(subset)}"'
        )

    subset = utils.make_iterable(subset)

    # Convert mask to vertex indices
    if subset.dtype == bool:
        if subset.shape[0] == x.vertices.shape[0]:
            subset = np.arange(len(x.vertices))[subset]
        elif subset.shape[0] == x.faces.shape[0]:
            # Translate face mask to vertex indices
            subset = np.unique(x.faces[subset])
        else:
            raise ValueError(
                "Boolean mask must be of same length as vertices or faces."
            )

    if prevent_fragments:
        # Generate skeleton
        sk = x.skeleton
        # Convert vertex IDs to node IDs
        subset_nodes = np.unique(x.skeleton.vertex_map[subset])
        # Find connected subgraph
        subset, _ = graph.connected_subgraph(x.skeleton, subset_nodes)
        # Convert node IDs back to vertex IDs
        subset = np.arange(0, len(x.vertices))[np.isin(sk.vertex_map, subset)]

    # Filter connectors
    # (connectors are associated with vertices, not faces which is why
    # our `subset` is always a list of vertex indices)
    if not keep_disc_cn and x.has_connectors:
        if "vertex_id" not in x.connectors.columns:
            x.connectors["vertex"] = x.snap(x.connectors[["x", "y", "z"]].values)[0]

        x._connectors = x.connectors[x.connectors.vertex.isin(subset)].copy()
        x._connectors.reset_index(inplace=True, drop=True)

        # Make old -> new indices map
        new_ix = dict(zip(subset, np.arange(0, len(subset))))

        x.connectors["vertex"] = x.connectors.vertex.map(new_ix)

    if len(subset):
        x.vertices, x.faces = submesh(x, vertex_index=subset)
    else:
        x.vertices, x.faces = np.empty((0, 3)), np.empty((0, 3))

    return x


def _subset_treeneuron(x, subset, keep_disc_cn, prevent_fragments):
    """Subset skeleton."""
    if isinstance(subset, (nx.DiGraph, nx.Graph)):
        subset = subset.nodes
    elif isinstance(subset, pd.DataFrame):
        subset = subset.node_id.values
    elif utils.is_iterable(subset):
        # This forces subset into numpy array (important for e.g. sets)
        subset = utils.make_iterable(subset)
    else:
        raise TypeError(
            "Can only subset to list, set, numpy.ndarray or"
            f'networkx.Graph, not "{type(subset)}"'
        )

    if prevent_fragments:
        subset, new_root = graph.connected_subgraph(x, subset)
    else:
        new_root = None  # type: ignore # new_root has already type from before

    # Filter nodes
    # Note that we are setting the nodes directly (here and later) thereby
    # circumventing (or rather postponing) checks and safeguards.
    if isinstance(subset, np.ndarray) and subset.dtype == bool:
        # For boolean mask
        x._nodes = x._nodes.loc[subset]
    else:
        # For sets of nodes
        x._nodes = x.nodes[x.nodes.node_id.isin(subset)]

    # Make sure that there are root nodes
    # This is the fastest "pandorable" way: instead of overwriting the column,
    # concatenate a new column to this DataFrame
    x._nodes = pd.concat(
        [
            x.nodes.drop("parent_id", inplace=False, axis=1),  # type: ignore  # no stubs for concat
            x.nodes[["parent_id"]].where(
                x.nodes.parent_id.isin(x.nodes.node_id.values), other=-1, inplace=False
            ),
        ],
        axis=1,
    )

    # Make sure any new roots or leafs are properly typed
    # We won't produce new slabs but roots, branches and leaves might change
    graph.classify_nodes(x, inplace=True)

    # Filter connectors
    if not keep_disc_cn and x.has_connectors:
        x._connectors = x.connectors[x.connectors.node_id.isin(x.nodes.node_id)]
        x._connectors.reset_index(inplace=True, drop=True)

    if getattr(x, "tags", None) is not None:
        # Filter tags
        x.tags = {
            t: [tn for tn in x.tags[t] if tn in x.nodes.node_id.values] for t in x.tags
        }  # type: ignore  # TreeNeuron has no tags

        # Remove empty tags
        x.tags = {t: x.tags[t] for t in x.tags if x.tags[t]}  # type: ignore  # TreeNeuron has no tags

    # Fix graph representations (avoids having to recompute them)
    if "_graph_nx" in x.__dict__:
        x._graph_nx = x.graph.subgraph(x.nodes.node_id.values)
    if "_igraph" in x.__dict__:
        if x.igraph and config.use_igraph:
            id2ix = {
                n: ix
                for ix, n in zip(
                    x.igraph.vs.indices, x.igraph.vs.get_attribute_values("node_id")
                )
            }
            indices = [id2ix[n] for n in x.nodes.node_id.values]
            vs = x.igraph.vs[indices]
            x._igraph = x.igraph.subgraph(vs)

    if hasattr(x, "_soma") and x._soma is not None:
        # Check if soma is still in the neuron
        if x._soma not in x.nodes.node_id.values:
            x._soma = None

    # Reset indices of data tables
    x.nodes.reset_index(inplace=True, drop=True)

    if new_root:
        x.reroot(new_root, inplace=True)

    return x


def submesh(mesh, *, faces_index=None, vertex_index=None):
    """Re-imlementation of trimesh.submesh that is faster for our use case.

    Notably we:
     - ignore normals (possibly needed) and visuals (definitely not needed)
     - allow only one set of faces to be passed
     - return vertices and faces instead of a new mesh
     - make as few copies as possible
     - allow passing vertex indices instead of faces

    This function is 5-10x faster than trimesh.submesh for our use case.
    Note that the speed of this function was never the bottleneck though,
    it's about the memory footprint.
    See https://github.com/navis-org/navis/issues/154.

    Parameters
    ----------
    mesh :          trimesh.Trimesh
                    Mesh to submesh.
    faces_index :   array-like
                    Indices of faces to keep.
    vertex_index :  array-like
                    Indices of vertices to keep.

    Returns
    -------
    vertices :  np.ndarray
                Vertices of submesh.
    faces :     np.ndarray
                Faces of submesh.

    """
    if faces_index is None and vertex_index is None:
        raise ValueError("Either `faces_index` or `vertex_index` must be provided.")
    elif faces_index is not None and vertex_index is not None:
        raise ValueError("Only one of `faces_index` or `vertex_index` can be provided.")

    # First check if we can return either an empty mesh or the original mesh right away
    if faces_index is not None:
        if len(faces_index) == 0:
            return np.array([]), np.array([])
        elif len(faces_index) == len(mesh.faces):
            if len(np.unique(faces_index)) == len(mesh.faces):
                return mesh.vertices.copy(), mesh.faces.copy()
    else:
        if len(vertex_index) == 0:
            return np.array([]), np.array([])
        elif len(vertex_index) == len(mesh.vertices):
            if len(np.unique(vertex_index)) == len(mesh.vertices):
                return mesh.vertices.copy(), mesh.faces.copy()

    # Use a view of the original data
    original_faces = mesh.faces.view(np.ndarray)
    original_vertices = mesh.vertices.view(np.ndarray)

    # If we're starting with vertices, find faces that contain at least one of our vertices
    # This way we will also make sure to drop unreferenced vertices
    if vertex_index is not None:
        faces_index = np.arange(len(original_faces))[
            np.isin(original_faces, vertex_index).all(axis=1)
        ]

    # Get unique vertices in the to-be-kept faces
    faces = original_faces[faces_index]
    unique = np.unique(faces.reshape(-1))

    # Generate a mask for the vertices
    # (using int32 here since we're unlikey to have more than 2B vertices)
    mask = np.arange(len(original_vertices), dtype=np.int32)

    # Remap the vertices to the new indices
    mask[unique] = np.arange(len(unique))

    # Grab the vertices in the order they are referenced
    vertices = original_vertices[unique].copy()

    # Remap the faces to the new vertex indices
    # (making a copy to allow `mask` to be garbage collected)
    faces = mask[faces].copy()

    return vertices, faces
